import os
from pathlib import Path
from typing import Any, Literal

import imandrax_api
import structlog
from imandrax_api import AsyncClient
from imandrax_api.twirp.exceptions import TwirpServerException

from .proto_models import (
    DecomposeRes,
    Error,
    ErrorMessage,
    EvalRes,
    InstanceRes,
    TypecheckRes,
    VerifyRes,
)
from .proto_to_dict import proto_to_dict

logger = structlog.get_logger(__name__)


def get_imandrax_client(
    imandra_api_key: str | None = None,
    env: Literal['dev', 'prod'] | None = None,
) -> AsyncClient:
    imandrax_env = env or os.getenv('IMANDRAX_ENV', 'prod')
    if imandrax_env == 'dev':
        url = imandrax_api.url_dev
    else:
        url = imandrax_api.url_prod

    imandrax_api_key = imandra_api_key or os.getenv('IMANDRAX_API_KEY')

    if not imandrax_api_key:
        # try to read from default config location
        config_path = Path.home() / '.config' / 'imandrax' / 'api_key'
        if config_path.exists():
            imandrax_api_key = config_path.read_text().strip()

    if not imandrax_api_key:
        logger.error('IMANDRAX_API_KEY is None')
        raise ValueError('IMANDRAX_API_KEY is None')
    client = AsyncClient(url=url, auth_token=imandrax_api_key, timeout=300)
    logger.info('imandrax_client_initialized', env=imandrax_env)
    return client


async def eval_src(
    imx_client: AsyncClient,
    src: str,
    timeout: float | None = None,
) -> EvalRes:
    res = await imx_client.eval_src(src=src, timeout=timeout)  # pyright: ignore
    return EvalRes.model_validate(proto_to_dict(res))


async def typecheck(
    imx_client: AsyncClient, src: str, timeout: float | None = None
) -> TypecheckRes:
    res = await imx_client.typecheck(src=src, timeout=timeout)  # pyright: ignore
    return TypecheckRes.model_validate(proto_to_dict(res))


async def decompose(
    imx_client: AsyncClient,
    name: str,
    assuming: str | None = None,
    basis: list[str] | None = None,
    rule_specs: list[str] | None = None,
    prune: bool | None = True,
    ctx_simp: bool | None = None,
    lift_bool: Any | None = None,
    timeout: float | None = None,
    str: bool | None = True,
) -> DecomposeRes:
    if basis is None:
        basis = []
    if rule_specs is None:
        rule_specs = []

    res = await imx_client.decompose(  # pyright: ignore
        name=name,
        basis=basis,
        rule_specs=rule_specs,
        prune=prune,
        ctx_simp=ctx_simp,
        lift_bool=lift_bool,
        timeout=timeout,
        str=str,
    )
    return DecomposeRes.model_validate(proto_to_dict(res))


async def verify_src(
    imx_client: AsyncClient,
    src: str,
    hints: str | None = None,
    timeout: float | None = None,
) -> VerifyRes:
    res = await imx_client.verify_src(src=src, hints=hints, timeout=timeout)  # pyright: ignore
    return VerifyRes.model_validate(proto_to_dict(res))


async def verify_src_catching_internal_error(
    imx_client: AsyncClient,
    src: str,
    hints: str | None = None,
    timeout: float | None = None,
) -> VerifyRes:
    try:
        return await verify_src(imx_client, src, hints, timeout)
    except TwirpServerException as e:
        if e.message == 'Internal Server Error':
            return VerifyRes(
                errors=[
                    Error(
                        msg=ErrorMessage(msg=e.to_dict()['meta']['body']),
                        kind='verify_internal',
                    )
                ]
            )
        raise e


async def instance_src(
    imx_client: AsyncClient,
    src: str,
    hints: str | None = None,
    timeout: float | None = None,
) -> InstanceRes:
    res = await imx_client.instance_src(src=src, hints=hints, timeout=timeout)  # pyright: ignore
    return InstanceRes.model_validate(proto_to_dict(res))



#
#   Imandra Inc.
#
#   model_cmds.py
#

import json
from typing import _LiteralGenericAlias, cast  # pyright: ignore

import imandra.u.agents.code_logician.command as cmd
from textual import on
from textual.binding import Binding
from textual.containers import HorizontalGroup, Vertical, VerticalGroup, VerticalScroll
from textual.message import Message
from textual.reactive import reactive
from textual.widgets import (
    Button,
    Input,
    Label,
    ListView,
    Markdown,
    Pretty,
    Select,
    Static,
)

from codelogician.modeling.model import Model
from codelogician.util import fst, guard, maybe_else

from ..common import Border


class AddCommand(Message):
    def __init__(self, command):
        self.command = command
        super().__init__()


class SelectIndex(Message):
    def __init__(self, index):
        self.index = index
        super().__init__()


class CommandForm(VerticalGroup):
    kind: reactive[None | type[cmd.BaseCommand]] = reactive(None, recompose=True)

    DEFAULT_CSS = """
        #field { padding-left: 1 }
        #buttons { width: auto }
        #fields { padding-left: 1; padding-right: 1; }
        Markdown { margin-left: 0; margin-top: 1; height: auto }
        Button { margin-left: 1; }
        HorizontalGroup { margin-left: 1; margin-bottom: 0}
        VerticalScroll#md { height: auto; max-height: 40%; background: $boost }
        VerticalScroll#fields { background: $boost; margin: 1 0 1 0 }
        Input.invalid { border: tall $error 70%; }
        Input.invalid:focus { border: tall $error; }
        #command-result { display: none }
        #command-result.filled { display: block }
        #errors { height: 6; width: 1fr; padding-left: 1; margin-right: 1; margin-left: 1 }
        #errors.valid { border: round $success; }
        #errors.invalid { border: round $error; }
        """

    def compose(self):
        def cmd_editor(c):
            import textwrap

            from pydantic_core import PydanticUndefined
            from rich.syntax import Syntax
            from rich.text import Text

            # fmt: off
            pl = {int: 'Integer', str: 'String', float: 'Decimal number'}
            def str_of_type(t): return t.__name__ if isinstance(t, type) else str(t)
            def not_undef(x): return x != PydanticUndefined
            def field_placeholder(f):
                return maybe_else(pl.get(f.annotation, ''), repr, guard(not_undef, f.default))
            # fmt: on

            fields = [(n, f) for n, f in c.model_fields.items() if n != 'type']

            with VerticalScroll(id='md'):
                # with Collapsible(title='Description', collapsed=True):
                # yield Markdown(textwrap.dedent(c.__doc__).strip('\n'))
                yield Markdown(textwrap.dedent(c.__doc__))

            if fields == []:
                yield Static('  (This command has no parameters.)')
            else:
                yield Static('')
                yield Label(
                    '  [b][$primary]Parameters (please enter in JSON syntax)[/][/b]'
                )
                with VerticalScroll(id='fields'):
                    for name, field in fields:
                        with Border(Text(name, 'bold')):
                            yield Markdown(field.description or 'No description')
                            with VerticalGroup(id='field'):
                                # fmt: off
                                with HorizontalGroup():
                                    req = 'required' if field.is_required() else 'optional'
                                    yield Label(Syntax(str_of_type(field.annotation), 'Python'))
                                    yield Label(Text(f' ({req})', 'dim'))
                                if type(field.annotation) is _LiteralGenericAlias:
                                    yield Select( [(v,v) for v in field.annotation.__args__], id=name)
                                else:
                                    placeholder = field_placeholder(field)
                                    yield Input(placeholder=placeholder, id=name)
                                # fmt: on
            with HorizontalGroup():
                with VerticalGroup(id='buttons'):
                    yield Button('Add', variant='primary', id='add')
                with VerticalScroll(id='errors'):
                    yield Pretty('', id='command-result')

        if self.kind != None:  # noqa
            yield from cmd_editor(self.kind)

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed):
        event.stop()

    # @on(Input.Submitted)
    # def show_invalid_reasons(self, event: Input.Submitted):
    #     result = event.validation_result
    #     reasons = [] if result.is_valid else result.failure_descriptions
    #     # self.query_one('#errors').update(repr(reasons))

    @on(Button.Pressed)
    def build_command(self, event):
        from pydantic import ValidationError

        # fmt: off
        def input_(n): return self.query_one('#' + n, Input | Select)  # pyright: ignore
        def json_loads(v):
            try: return json.loads(v)
            except Exception:
                raise
        def bare_string(t):
            return t in [str, bool] or type(t) is _LiteralGenericAlias
        def decode(t, v):
            return (None if v == '' else
                    v if bare_string(t) else
                    json.loads(v))

        c = self.kind
        if c is None: return

        input_by_name = {n: (input_(n), f.annotation)
                         for n, f in c.model_fields.items() if n != 'type'}
        # fmt: on
        w = self.query_one('#errors')
        r = self.query_one('#command-result', Pretty)
        r.add_class('filled')
        for i, _ in input_by_name.values():
            i.remove_class('invalid')
        try:
            name_val_pairs = [
                (n, decode(t, i.value)) for n, (i, t) in input_by_name.items()
            ]
            dict_val = {n: v for n, v in name_val_pairs if v is not None}
            command = c.model_validate(dict_val)
            r.update(command)
            w.border_title = 'Valid'
            w.remove_class('invalid')
            w.add_class('valid')
            if event.button.id == 'add':
                self.screen.query_one(ModelCommandsView).post_message(
                    AddCommand(command)
                )
        except json.JSONDecodeError as e:
            r.update(e)
            w.border_title = 'JSON Decode Error'
            w.remove_class('valid')
            w.add_class('invalid')
        except TypeError as e:
            r.update(e)
            w.border_title = 'Type error'
            w.remove_class('valid')
            w.add_class('invalid')

        except ValidationError as e:
            r.update(e)
            for err in e.errors():
                for n in err['loc']:
                    if isinstance(n, str):
                        fst(input_by_name[n]).add_class('invalid')

            w.border_title = 'Error'
            w.remove_class('valid')
            w.add_class('invalid')


class CommandEditor(Vertical):
    DEFAULT_CSS = 'Select { margin-bottom: 1 } CommandForm { height: 1fr }'

    def compose(self):
        command_classes = [
            cmd.AdmitModelCommand,
            cmd.AgentFormalizerCommand,
            cmd.CheckFormalizationCommand,
            cmd.EditStateElementCommand,
            cmd.GenFormalizationDataCommand,
            cmd.GenFormalizationFailureDataCommand,
            cmd.GenModelCommand,
            cmd.GenProgramRefactorCommand,
            cmd.GenRegionDecompsCommand,
            cmd.GenTestCasesCommand,
            cmd.GenVgsCommand,
            cmd.GetStateElementCommand,
            cmd.InitStateCommand,
            cmd.InjectFormalizationContextCommand,
            cmd.SearchFDBCommand,
            cmd.SetModelCommand,
            cmd.SuggestApproximationCommand,
            cmd.SuggestAssumptionsCommand,
            cmd.SuggestFormalizationActionCommand,
            cmd.SyncModelCommand,
            cmd.SyncSourceCommand,
        ]

        yield Select([(c.__name__, c) for c in command_classes])
        yield CommandForm(id='editor')

    @on(Select.Changed)
    def select_changed(self, event: Select.Changed):
        editor = self.query_one('#editor', CommandForm)
        editor.kind = guard(
            lambda x: x != Select.BLANK, cast(type[cmd.BaseCommand], event.value)
        )
        # editor.kind = event.value


# fmt: off
def move_up(i, x):   return (i-1, [*x[:i-1], x[i], x[i-1], *x[i+1:]]) if i > 0        else (i, x)
def move_down(i, x): return (i+1, [*x[:i], x[i+1], x[i], *x[i+2:]])   if i < len(x)-1 else (i, x)
def rm_nth(i, x):    return (None, [*x[:i], *x[i+1:]])
def maybe_(f, i, x): return (i, x) if i is None else f(i, x)
# fmt: on


class CommandsList(Vertical):
    DEFAULT_CSS = """
        HorizontalGroup { height: 3; }
        Button { border: none; min-width: 1; }
        View { height: 90%}
    """

    class View(ListView):
        commands = reactive([], recompose=True)

        BINDINGS = [
            Binding('backspace', 'delete', 'Delete', show=False),
            Binding('shift+up', 'move_up', 'Move up', show=False),
            Binding('shift+down', 'move_down', 'Move down', show=False),
        ]

        def __init__(self, owner):
            ListView.__init__(self)
            self.owner = owner

        def action_delete(self):
            self.index, self.commands = maybe_(rm_nth, self.index, self.commands)

        def action_move_up(self):
            self.index, self.commands = maybe_(move_up, self.index, self.commands)

        def action_move_down(self):
            self.index, self.commands = maybe_(move_down, self.index, self.commands)

        def compose(self):
            from textual.widgets import ListItem

            for c in self.commands:
                yield ListItem(Static(c.__rich__()))

        def watch_index(self, old, new):
            ListView.watch_index(self, old, new)
            self.owner.post_message(SelectIndex(new))

    def on_mount(self):
        self.lv = self.query_one(CommandsList.View)

    def compose(self):
        # from textual.containers import Grid
        with HorizontalGroup():
            yield Button('move up', id='up', disabled=True)
            yield Button('move down', id='down', disabled=True)
            yield Button('delete', id='del', disabled=True)
        yield CommandsList.View(self)

    @on(SelectIndex)
    def selected_index(self, event):
        for b in self.query(Button):
            b.disabled = event.index is None

    @on(Button.Pressed)
    def edit(self, event):
        {
            'del': self.lv.action_delete,
            'up': self.lv.action_move_up,
            'down': self.lv.action_move_down,
        }[event.button.id]()

    def add_command(self, command):
        v = self.query_one(CommandsList.View)
        v.commands = [*v.commands, command]


class ModelCommandsView(VerticalScroll):
    DEFAULT_CSS = 'CommandsList { height: 40% }'

    def __init__(self, model: Model, container):
        super().__init__()
        self.styles.height = 'auto'
        self.model = model
        self.container = container

    def compose(self):
        # from textual.containers import Grid

        with Vertical():
            yield CommandEditor()
            yield CommandsList(id='commands')

    @on(AddCommand)
    def add_command(self, event):
        commands = self.query_one('#commands', CommandsList)
        commands.add_command(event.command)



"""Utility functions - Level 3 (depends on multiple packages)"""

from core.advanced import factorial
from core.arithmetic import add, divide, multiply
from stats.basic import mean
from stats.measures import standard_deviation


def polynomial_evaluate(coefficients: list[float], x: float) -> float:
    """Evaluate polynomial at x using Horner's method."""
    if not coefficients:
        return 0.0

    result = coefficients[0]
    for i in range(1, len(coefficients)):
        result = add(multiply(result, x), coefficients[i])

    return result


def combinations(n: int, r: int) -> int:
    """Calculate combinations C(n,r)."""
    if r < 0 or r > n or n < 0:
        return 0

    if r == 0 or r == n:
        return 1

    return int(divide(factorial(n), multiply(factorial(r), factorial(n - r))))


def normalize_data(data: list[float]) -> list[float]:
    """Normalize data to have mean 0 and std 1."""
    if not data:
        return []

    data_mean = mean(data)
    data_std = standard_deviation(data)

    if data_std == 0:
        return [0.0] * len(data)

    normalized = []
    for value in data:
        normalized_value = divide(add(value, multiply(-1.0, data_mean)), data_std)
        normalized.append(normalized_value)

    return normalized



# ruff: noqa: F841
from __future__ import annotations

import base64
from typing import Any

from imandrax_api.lib import (
    Artifact,
    Common_Applied_symbol_t_poly,
    Common_Fun_decomp_t_poly,
    Common_Model_fi,
    Common_Model_t_poly,
    Common_Model_ty_def,
    Common_Region_t_poly,
    Common_Var_t_poly,
    Mir_Fun_decomp,
    Mir_Model,
    Mir_Term,
    Mir_Term_view_Const,
    Mir_Term_view_Construct,
    Mir_Type,
    Uid,
    read_artifact_data,
)
from pydantic import BaseModel, Field


class RegionStr(BaseModel):
    constraints_str: list[str] | None = Field(default=None)
    invariant_str: str
    model_str: dict[str, str]
    model_eval_str: str | None


class ArtifactDecodeError(Exception):
    pass


def to_bytes(x: bytes | str) -> bytes:
    return x if isinstance(x, bytes) else base64.b64decode(x)


def decode_artifact(
    data: bytes | str,
    kind: str,
) -> list[RegionStr] | dict[str, str] | None:
    art: Artifact = read_artifact_data(data=to_bytes(data), kind=kind)
    match (art, kind):
        # DecomposeRes
        case (
            Common_Fun_decomp_t_poly(
                f_id=_f_id,
                f_args=_f_args,
                regions=regions,
            ) as _fun_decomp,
            'mir.fun_decomp',
        ):
            _fun_decomp: Mir_Fun_decomp
            _f_id: Uid
            _f_args: list[Common_Var_t_poly[Mir_Type]]
            regions: list[Common_Region_t_poly[Mir_Term, Mir_Type]]

            return unwrap_region_str(regions)
        # VerifyRes / InstanceRes
        case (
            Common_Model_t_poly(
                tys=tys,
                consts=consts,
                funs=funs,
                representable=representable,
                completed=completed,
                ty_subst=ty_subst,
            ) as mir_model,
            'mir.model',
        ):
            mir_model: Mir_Model
            tys: list[tuple[Mir_Type, Common_Model_ty_def[Mir_Term, Mir_Type]]]
            consts: list[tuple[Common_Applied_symbol_t_poly[Mir_Type], Mir_Term]]
            funs: list[
                tuple[
                    Common_Applied_symbol_t_poly[Mir_Type],
                    Common_Model_fi[Mir_Term, Mir_Type],
                ]
            ]
            representable: bool
            completed: bool
            ty_subst: list[tuple[Uid, Mir_Type]]

            consts_d: dict[str, Any] = unwrap_model_constants(consts)
            return consts_d
        case _:
            raise ArtifactDecodeError(
                f'Unknown artifact type: {type(art)}, with {kind = }'
            )


def unwrap_model_constants(
    consts: list[tuple[Common_Applied_symbol_t_poly[Mir_Type], Mir_Term]],
) -> dict[str, Any]:
    constants: dict[str, Any] = {}

    for applied_symbol, term in consts:
        match applied_symbol, term:
            case (
                Common_Applied_symbol_t_poly(
                    sym=applied_symbol_sym,
                    args=applied_symbol_args,
                    ty=applied_symbol_ty,
                ),
                Mir_Term(view=term_view, ty=term_ty, sub_anchor=term_sub_anchor),
            ):
                var_name = applied_symbol_sym.id.name

                # Extract the value from the term
                match term_view:
                    case Mir_Term_view_Const(arg=term_view_const):
                        constants[var_name] = term_view_const.arg
                    case Mir_Term_view_Construct():
                        raise NotImplementedError(
                            'Term view type of Mir_Term_view_Construct is not supported'
                        )
                    case _:
                        raise ValueError(
                            f'Unexpected term view type: {type(term_view)}'
                        )

    return constants


def unwrap_region_str(
    regions: list[Common_Region_t_poly[Mir_Term, Mir_Type]],
) -> list[RegionStr]:
    """
    A region object looks like:
        {
            "constraints": [...],
            "invariant": ...,
            "meta": [
                ("str", ...)  # What we want
                ("model_eval", ...)
                ("id", "...")
            ]
            "status": ...
        }
    """
    regions_str: list[RegionStr] = []
    for region in regions:
        match region:
            case Common_Region_t_poly(
                constraints=_constraints,
                invariant=_invariant,
                meta=meta,
                status=_status,
            ):
                meta_str_d: dict[str, Any] = dict(dict(meta)['str'].arg)  # pyright: ignore
                constraints: list[str] = [c.arg for c in meta_str_d['constraints'].arg]
                invariant: str = meta_str_d['invariant'].arg
                model: dict[str, str] = {k: v.arg for (k, v) in meta_str_d['model'].arg}

                model_eval: str | None
                if 'model_eval' in meta_str_d:
                    model_eval = meta_str_d['model_eval'].arg
                else:
                    model_eval = None

                #
                region_str = RegionStr(
                    invariant_str=invariant,
                    constraints_str=constraints,
                    model_str=model,
                    model_eval_str=model_eval,
                )
                regions_str.append(region_str)
            case _:
                raise ValueError(f'Unknown region type: {type(region)}')
    return regions_str



#
#   Imandra Inc.
#
#   codelogician/commands/cmd_multiagent.py
#

from typing import Annotated

import typer

from codelogician.util import setup_logging


# fmt: off
def run_multiagent_cmd(
    dir   : Annotated[str, typer.Argument(help='Target directory')],
    clean : Annotated[bool, typer.Option(help='Start clean by disregarding any existing cache files')] = False,
    config: Annotated[str, typer.Option(help='Server configuration YAML file')] = 'config/server_config.yaml',
    debug : Annotated[bool, typer.Option(help='Debug mode')] = False,
):
    from codelogician.server.oneshot import run_oneshot

    setup_logging(debug=debug)
    run_oneshot(dir, clean, config)
# fmt: on



"""Numerical derivative calculations - Level 4"""

from core.arithmetic import add, divide, multiply, subtract


def numerical_derivative(func, x: float, h: float = 1e-8) -> float:
    """Calculate numerical derivative using central difference."""
    x_plus_h = add(x, h)
    x_minus_h = subtract(x, h)

    f_plus = func(x_plus_h)
    f_minus = func(x_minus_h)

    numerator = subtract(f_plus, f_minus)
    denominator = multiply(2.0, h)

    return divide(numerator, denominator)


def second_derivative(func, x: float, h: float = 1e-6) -> float:
    """Calculate second derivative."""
    x_plus_h = add(x, h)
    x_minus_h = subtract(x, h)

    f_center = func(x)
    f_plus = func(x_plus_h)
    f_minus = func(x_minus_h)

    # f''(x) ≈ (f(x+h) - 2f(x) + f(x-h)) / h²
    numerator = subtract(add(f_plus, f_minus), multiply(2.0, f_center))
    denominator = multiply(h, h)

    return divide(numerator, denominator)



#
#   Imandra Inc.
#
#   fs_sync.py
#

import logging
from pathlib import Path
from typing import cast

from ..dep_tools.python import build_py_dep_graph
from ..tools.filesystem import FileSystem
from .model import Model

log = logging.getLogger(__name__)


def model_filesystem_sync(
    models: dict[str, Model],
    src_dir_abs_path: str,
    language: str,
) -> dict[str, Model]:
    """
    Go through the source code directory and update the models' dependencies/create new ones if needed.
    """

    fs = FileSystem.from_disk(Path(src_dir_abs_path))
    py_dep = build_py_dep_graph(fs)

    # Let's look for deleted nodes
    model_paths = {str(m.relative_path) for m in py_dep.nx_graph}
    for path in list(models.keys()):
        log.info(f'Checking path {path}...')
        if path not in model_paths:
            models[path].set_src_code(None)

    # Now let's look at new nodes / update code for the existing ones
    for node in py_dep.nodes:
        rel_path = str(node.relative_path)
        if rel_path not in models:
            log.info(f'Adding model for path {rel_path}')
            models[rel_path] = Model(
                rel_path=rel_path,
                src_code=node.content,
                src_language=language,
            )
            log.info(f'Creating new model with path={node.relative_path}.')
        else:
            log.info(f'Setting source code to existing node={node.relative_path}.')
            models[rel_path].set_src_code(node.content)

    # Rebuild edges from scratch
    for m in models.values():
        m.dependencies.clear()
        m.rev_dependencies.clear()

    # Now let's add dependencies + reverse dependencies
    for p1, p2 in py_dep.nx_graph.edges:
        path1, path2 = str(p1.relative_path), str(p2.relative_path)

        if path1 not in models or path2 not in models:
            errorMsg = "Attempting to add a dependency that doesn't exist!"
            log.error(errorMsg)
            raise Exception(errorMsg)

        model1 = models[path1]  # depends-on
        model2 = models[path2]  # dependency

        if model2 not in cast(list[Model], model1.dependencies):
            model1.dependencies.append(model2)

        if model1 not in cast(list[Model], model2.rev_dependencies):
            model2.rev_dependencies.append(model1)

    return models



#
#   Imandra Inc.
#
#   pyiml_strategy.py
#

from __future__ import annotations

import logging
from queue import Empty, Queue
from threading import Thread

from rich.progress import Progress, TaskID

from ..server.events import ServerEvent
from .config import StratConfig
from .events import (
    AutoModeCLTaskEvent,
    CLResultEvent,
    StrategyEvent,
)
from .pyiml_processor import PyIMLStrategyProcessor

log = logging.getLogger(__name__)


class PyIMLStrategy(Thread):
    """
    Strategy thread wrapper.

    The actual behavior is implemented in `PyIMLStrategyProcessor` for testability.
    """

    def __init__(
        self,
        state,
        config: StratConfig,
        oneshot: bool = False,
    ):
        super().__init__(daemon=True)

        # Processor owns all logic/stateful collaborators (worker, timers, etc.)
        self._proc = PyIMLStrategyProcessor(state=state, config=config, oneshot=oneshot)

        # This will be our events queue (we use `None` to stop the thread)
        self._queue: Queue[StrategyEvent | ServerEvent | None] = Queue()

        # Wire processor's callback to enqueue events
        self._proc.set_emit(self.add_event)

        log.info('Started PyIMLStrategy thread wrapper')

    # -------------------------
    # Passthroughs (public API)
    # -------------------------

    def language(self) -> str:
        return self._proc.language()

    def extensions(self) -> list[str]:
        return self._proc.extensions()

    def config(self):
        return self._proc.config

    def update_config(self, new_config) -> None:
        self._proc.update_config(new_config)

    def state(self):
        return self._proc.state

    def add_event(self, event: StrategyEvent | ServerEvent | None) -> None:
        """
        Add event to the queue for processing (None will force shutdown)
        """
        self._queue.put(event)
        log.info(
            f'Event added: {type(event).__name__ if event is not None else "None"}'
        )

    # -------------------------
    # Thread loop
    # -------------------------

    def run(self) -> None:
        # Let processor initialize (worker thread, initial sync, etc.)
        self._proc.start()

        if self._proc.oneshot:
            self._run_oneshot()
            return

        while True:
            try:
                event = self._queue.get()  # Blocks until an item is available

                if event is None:
                    log.info("Received a 'None' event to process. Shutting down.")
                    self._proc.shutdown()
                    break

                self._proc.process_event(event)
                self._queue.task_done()

            except Empty:
                pass
            except Exception as e:
                log.error(f'Caught exception during event processing: {str(e)}')

    def _run_oneshot(self) -> None:
        """
        One-shot loop. (Progress UI stays here; the business logic stays in processor.)
        """

        # We'll keep the progress of the various tasks here...
        progressTasks: dict[str, TaskID] = {}

        mmodel = self._proc.state.curr_meta_model
        if not mmodel:
            log.warning('Oneshot requested but no current MetaModel. Shutting down.')
            self._proc.shutdown()
            return

        with Progress() as progress:
            for path in mmodel.models:
                progressTasks[path] = progress.add_task(
                    f'[green] Formalizing {path}', total=1000
                )

            while True:
                try:
                    event = self._queue.get()

                    if event is None:
                        log.info("Received a 'None' event to process. Shutting down.")
                        self._proc.shutdown()
                        break

                    # Progress updates based on event type
                    if isinstance(event, AutoModeCLTaskEvent):
                        progress.update(
                            task_id=progressTasks[event.task.rel_path], completed=250
                        )
                    elif isinstance(event, CLResultEvent):
                        progress.update(
                            task_id=progressTasks[event.result.task.rel_path],
                            completed=1000,
                        )

                    self._proc.process_event(event)
                    self._queue.task_done()

                except Empty:
                    pass
                except Exception as e:
                    log.error(f'Caught exception during event processing: {str(e)}')



#
#   Imandra Inc.
#
#   codelogician/modeling/dep_context.py
#

from __future__ import annotations

from collections import deque
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import cast

from imandra.u.agents.code_logician.base import FormalizationDependency, ModuleInfo

from ..util import filter as cl_filter, map as cl_map
from .model import Model


@dataclass(frozen=True)
class CycleEdge:
    src: Model
    dst: Model


@dataclass(frozen=True)
class CycleInfo:
    """
    A concrete dependency cycle.

    - nodes is a closed walk: nodes[0] == nodes[-1]
    - edges align with nodes: edges[i] is nodes[i] -> nodes[i+1]
    """

    nodes: list[Model]
    edges: list[CycleEdge]

    @property
    def rel_paths(self) -> list[str]:
        return [m.rel_path for m in self.nodes]

    @property
    def edge_paths(self) -> list[tuple[str, str]]:
        return [(e.src.rel_path, e.dst.rel_path) for e in self.edges]


@dataclass(frozen=True)
class ModelDepContext:
    """
    Dependency analysis utilities for Model objects.

    This isolates graph traversal, dependency-to-ModuleInfo conversion,
    and "deps changed" computations away from Model and task policy.
    """

    @staticmethod
    def deps(model: Model) -> list[Model]:
        # "live" deps are Model objects; deserialized ones may be strings
        if not model.dependencies or isinstance(model.dependencies[0], str):
            return []
        return cast(list[Model], model.dependencies)

    @staticmethod
    def deps_paths(model: Model) -> list[str]:
        return [d.rel_path for d in ModelDepContext.deps(model)]

    @staticmethod
    def no_module_deps(model: Model) -> bool:
        return not model.dependencies or isinstance(model.dependencies[0], str)

    # -------------------------
    # ModuleInfo conversion
    # -------------------------
    @staticmethod
    def make_dep_name_from_path(rel_path: str) -> str:
        p = Path(rel_path).parent / Path(rel_path).stem
        return ('_'.join(Path(p).parts)).capitalize()

    @staticmethod
    def to_src_module_info(model: Model) -> ModuleInfo:
        return ModuleInfo(
            name=ModelDepContext.make_dep_name_from_path(model.rel_path),
            relative_path=Path(model.rel_path),
            content=model.src_code or '',
            src_lang=model.src_language,
        )

    @staticmethod
    def to_iml_module_info(model: Model) -> ModuleInfo | None:
        iml = model.iml_code()
        if iml is None:
            return None
        return ModuleInfo(
            name=ModelDepContext.make_dep_name_from_path(model.rel_path),
            relative_path=Path(model.rel_path),
            content=str(iml),
            src_lang='IML',
        )

    # -------------------------
    # Graph traversal / sorting
    # -------------------------
    @staticmethod
    def full_deep_dependencies(model: Model) -> list[Model]:
        result: list[Model] = []

        def visit(m: Model) -> None:
            for d in ModelDepContext.deps(m):
                if d not in result:
                    result.append(d)
                visit(d)

        visit(model)
        return result

    @staticmethod
    def topological_sort(models: list[Model]) -> list[Model]:
        """
        Return list sorted topologically by dependencies.
        (Same algorithm you had, but moved here.)
        """
        path_to_idx: dict[str, int] = {}
        idx_to_path: dict[int, str] = {}
        edges: list[tuple[int, int]] = []
        idx = 0

        def add_to_lookup(i: int, path: str) -> None:
            path_to_idx[path] = i
            idx_to_path[i] = path

        for m in models:
            if m.rel_path not in path_to_idx:
                add_to_lookup(idx, m.rel_path)
                idx += 1

            for dep_path in ModelDepContext.deps_paths(m):
                if dep_path not in path_to_idx:
                    add_to_lookup(idx, dep_path)
                    idx += 1
                edges.append((path_to_idx[m.rel_path], path_to_idx[dep_path]))

        def construct_adj(v: int, es: list[tuple[int, int]]) -> list[list[int]]:
            adj = [[] for _ in range(v)]
            for u, w in es:
                adj[u].append(w)
            return adj

        def topo(v: int, es: list[tuple[int, int]]) -> list[int]:
            adj = construct_adj(v, es)
            indegree = [0] * v
            for u in range(v):
                for w in adj[u]:
                    indegree[w] += 1
            q = deque([i for i in range(v) if indegree[i] == 0])

            out: list[int] = []
            while q:
                node = q.popleft()
                out.append(node)
                for nb in adj[node]:
                    indegree[nb] -= 1
                    if indegree[nb] == 0:
                        q.append(nb)

            if len(out) != v:
                raise Exception('Graph contains cycle!')
            return out

        model_by_path = {m.rel_path: m for m in models}
        sorted_idxs = topo(len(models), edges)
        return [model_by_path[idx_to_path[i]] for i in reversed(sorted_idxs)]

    # -------------------------
    # Cycle detection (returns actual nodes + actual edges)
    # -------------------------
    @staticmethod
    def detect_dependency_cycles(model: Model) -> list[CycleInfo]:
        """
        Detect directed cycles reachable from `model` following `dependencies`.

        Returns concrete CycleInfo (nodes + edges).

        Notes:
        - Designed for "live" graphs (dependencies as Model objects).
        - If deps are not live (strings / empty), returns [].
        """
        if ModelDepContext.no_module_deps(model):
            return []

        WHITE, GRAY, BLACK = 0, 1, 2  # noqa
        color: dict[str, int] = {}
        parent: dict[str, str | None] = {}
        node_by_path: dict[str, Model] = {}
        cycles: list[CycleInfo] = []

        def reconstruct_cycle_paths(from_path: str, to_path: str) -> list[str]:
            # Cycle is: to_path -> ... -> from_path -> to_path
            path = [to_path]
            cur = from_path
            while cur != to_path and cur is not None:
                path.append(cur)
                cur = parent.get(cur)
            path.append(to_path)
            path.reverse()
            return path

        def record_cycle(from_path: str, to_path: str) -> None:
            path = reconstruct_cycle_paths(from_path, to_path)

            # Convert to concrete nodes
            nodes: list[Model] = []
            for p in path:
                m = node_by_path.get(p)
                if m is None:
                    return
                nodes.append(m)

            # Build concrete edges and sanity-check adjacency
            edges: list[CycleEdge] = []
            for i in range(len(nodes) - 1):
                src = nodes[i]
                dst = nodes[i + 1]
                if dst not in ModelDepContext.deps(src):
                    # Graph mutated during traversal or inconsistent parent pointers.
                    return
                edges.append(CycleEdge(src=src, dst=dst))

            # Closed walk invariant (best-effort)
            if nodes and nodes[0] is nodes[-1]:
                cycles.append(CycleInfo(nodes=nodes, edges=edges))
            elif nodes and nodes[0].rel_path == nodes[-1].rel_path:
                # Fallback if identity differs but rel_path is same (shouldn’t happen in one graph)
                cycles.append(CycleInfo(nodes=nodes, edges=edges))

        def dfs(m: Model) -> None:
            mp = m.rel_path
            node_by_path[mp] = m
            color[mp] = GRAY

            for d in ModelDepContext.deps(m):
                dp = d.rel_path
                node_by_path[dp] = d

                if dp not in color:
                    parent[dp] = mp
                    dfs(d)
                elif color[dp] == GRAY:
                    # back-edge => cycle
                    record_cycle(mp, dp)

            color[mp] = BLACK

        parent[model.rel_path] = None
        dfs(model)
        return cycles

    @staticmethod
    def detect_cycles_in_models(models: Iterable[Model]) -> list[CycleInfo]:
        """
        Detect cycles anywhere in the given model set (live graph).
        """
        out: list[CycleInfo] = []
        seen_roots: set[str] = set()

        for m in models:
            if m.rel_path in seen_roots:
                continue
            seen_roots.add(m.rel_path)
            out.extend(ModelDepContext.detect_dependency_cycles(m))

        return out

    # -------------------------
    # Dependency payloads for CL
    # -------------------------
    @staticmethod
    def shallow_dep_iml_models(model: Model) -> list[FormalizationDependency]:
        if ModelDepContext.no_module_deps(model):
            return []

        ds: list[FormalizationDependency] = []
        for dep in ModelDepContext.deps(model):
            src_mod = ModelDepContext.to_src_module_info(dep)
            iml_mod = ModelDepContext.to_iml_module_info(dep)
            if iml_mod is None:
                continue
            ds.append(FormalizationDependency(src_module=src_mod, iml_module=iml_mod))
        return ds

    @staticmethod
    def dependencies_iml_models(model: Model) -> list[FormalizationDependency]:
        """
        Deep dependency list, topologically ordered, including only deps that have IML.
        """
        if ModelDepContext.no_module_deps(model):
            return []

        full_list = ModelDepContext.full_deep_dependencies(model)
        sorted_list = ModelDepContext.topological_sort(full_list)

        ds: list[FormalizationDependency] = []
        for dep in sorted_list:
            src_mod = ModelDepContext.to_src_module_info(dep)
            iml_mod = ModelDepContext.to_iml_module_info(dep)
            if iml_mod is None:
                continue
            ds.append(FormalizationDependency(src_module=src_mod, iml_module=iml_mod))
        return ds

    @staticmethod
    def _dep_fingerprint(d: FormalizationDependency) -> tuple[str, str]:
        """
        Stable fingerprint for dependency-change detection.

        We intentionally DO NOT include src_module.content because:
        - deps_changed is used to decide whether the last formalization is still valid
        - the invariant we care about is that imported IML for deps hasn't changed
        - src_module.content can change independently and would cause perpetual re-formalization

        Fingerprint = (relative_path, iml_content)
        """
        rel = str(d.src_module.relative_path)
        iml = d.iml_module.content if d.iml_module is not None else ''
        return (rel, iml)

    @staticmethod
    def deps_changed(model: Model) -> bool:
        """
        Compare current shallow deps (with IML) against deps recorded at last formalization.
        """
        current = ModelDepContext.shallow_dep_iml_models(model)

        paths_list = cl_map(Path, ModelDepContext.deps_paths(model))
        in_paths = lambda x: x.src_module.relative_path in paths_list  # noqa: E731
        relevant_formalized = cl_filter(in_paths, model.formalized_deps)

        current_fp = {ModelDepContext._dep_fingerprint(d) for d in current}
        formalized_fp = {
            ModelDepContext._dep_fingerprint(d) for d in relevant_formalized
        }

        return current_fp != formalized_fp

    @staticmethod
    def create_import_iml_statements(model: Model) -> str:
        text = ''
        for d in ModelDepContext.dependencies_iml_models(model):
            name = d.iml_module.name
            rel_path = str(Path(d.iml_module.relative_path).with_suffix('.iml'))
            text += f'[@@@import {name}, "{rel_path}"]\n'
        return text

    @staticmethod
    def to_CL_src_module_info(m: Model) -> ModuleInfo:
        """
        Create ModuleInfo
        """
        return ModuleInfo(
            name=ModelDepContext.make_dep_name_from_path(m.rel_path),
            relative_path=Path(m.rel_path),
            content=m.src_code if m.src_code else '',
            src_lang=m.src_language,
        )

    @staticmethod
    def to_CL_iml_module_info(m: Model) -> ModuleInfo | None:
        """
        Return ModuleInfo object for the IML code if it's available, None otherwise
        """

        # If we don't have any IML code for this model, let's not return it
        if m.iml_code() is None:
            return None

        if m.iml_code() is not None:
            iml_code = str(m.iml_code())
        else:
            iml_code = ''

        return ModuleInfo(
            name=ModelDepContext.make_dep_name_from_path(m.rel_path),
            relative_path=Path(m.rel_path),
            content=iml_code,
            src_lang='IML',
        )



#
# Imandra Inc.
#
# endpts_strategy.py
#

# TODO need to implement this vs manually creating an pydantic model
# from pydantic_partial import create_partial_model
# PartialPyIMLConfig = create_partial_model(StratConfig)
import logging

from fastapi import HTTPException

from codelogician.strategy.config import StratConfig, StratConfigUpdate
from codelogician.strategy.state import StrategyState

from ..cl_server import CLServer

log = logging.getLogger(__name__)


def register_strategy_endpoints(app: CLServer):  # noqa: C901
    """
    Functions relate to accessing the state of the strategy or sending it commands
    """

    @app.post('/strategy/setcws/{strat_id}', operation_id='setcws')
    async def set_cws(strat_id: str) -> str:
        """
        Set the current working strategy
        """

        try:
            app.set_current_strat_id(strat_id)
        except Exception as e:
            raise HTTPException(
                status_code=400,
                detail=f'Failed to update current working strategy: {e}',
            )

        return 'OK'

    @app.get('/strategy/cws', operation_id='cws')
    async def get_cws() -> str | None:
        """
        Return ID of the current working strategy
        """
        return app.current_strat_id

    @app.get('/strategy/list', operation_id='list_strategies')
    async def list_strategies() -> list[dict[str, str]]:
        """
        Return the current list of strategies. Each strategy will contain:
        - 'strat_id' - strat_id
        - 'path' - directory path for this strategy
        - 'type' - 'PyIML' will be returned as it's the only strategy currently supported
        """
        return app.list_strategies()

    @app.get('/strategy/states', operation_id='get_all_strategy_states')
    async def all_strategy_states() -> list[StrategyState]:
        """
        Return all strategy states
        """
        return list(app.strategy_states().values())

    @app.post('/strategy/create', operation_id='create_new_strategy')
    async def create_strategy(strat_type: str, strat_path: str) -> str:
        """
        Create a new strategy
        """
        try:
            strat_id = app.add_strategy(strat_type, strat_path)
        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f'Failed to create strategy: {e}'
            )

        return strat_id

    @app.post('/strategy/delete/{strat_id}', operation_id='delete_strategy')
    async def delete_strategy(strat_id: str) -> str:
        """
        Remove strategy (will stop the worker and delete it from the container)
        """

        if app.current_strat_id is None:
            raise HTTPException(status_code=400, detail='No current strategy')

        try:
            app.rem_strategy(app.current_strat_id)
        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f'Failed to delete strategy: {str(e)}'
            )

        return 'OK'

    @app.get('/strategy/state', operation_id='get_strategy_state')
    async def strat_state() -> StrategyState:
        """
        Return the full strategy state
        """
        log.info('Received strategy state request')

        if app.current_strat_id is None:
            raise HTTPException(status_code=400, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(app.current_strat_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        return state

    @app.get('/strategy/summary', operation_id='get_strategy_summary')
    async def strat_summary() -> dict:
        """
        Return PyIML strategy state summary
        """
        log.info('Received request for strategy summary')

        if app.current_strat_id is None:
            raise HTTPException(status_code=400, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(app.current_strat_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        return state.summary()

    @app.get('/strategy/config', operation_id='get_strategy_config')
    async def strat_config() -> StratConfig:
        """
        Return the current strategy configuration
        """
        log.info('Received request for strategy config')

        if app.current_strat_id is None:
            raise HTTPException(status_code=400, detail='No current strategy')

        try:
            strat_config = app.strategy_config_by_id(app.current_strat_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        return strat_config

    @app.patch('/strategy/config/set', operation_id='set_config_field')
    async def strat_config_set(configUpdate: StratConfigUpdate) -> str:
        """
        Set strategy configuration field to specified value.
        """

        if app.current_strat_id is None:
            raise HTTPException(status_code=400, detail='No current strategy')

        try:
            strat_config = app.strategy_config_by_id(app.current_strat_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy config: {e}'
            )

        log.info(
            f'Received strategy config update request: {str(configUpdate.model_dump())}'
        )
        try:
            updated_data = configUpdate.model_dump(exclude_unset=True)
            for key, value in updated_data.items():
                setattr(strat_config, key, value)

            app.update_strat_config(app.current_strat_id, strat_config)
        except Exception as e:
            err_msg = f'Error when attempting to update config: {str(e)}'
            log.error(err_msg)
            raise HTTPException(status_code=406, detail=err_msg)

        return 'OK'



#
#   Imandra Inc.
#
#   model.py
#

# ruff: noqa: UP031
import sys
from pathlib import Path
from typing import cast

from textual import on
from textual.containers import (
    Horizontal,
    HorizontalGroup,
    ScrollableContainer,
    VerticalGroup,
    VerticalScroll,
)
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import (
    Button,
    Collapsible,
    Footer,
    Label,
    Static,
    TabbedContent,
    TabPane,
)

from codelogician.modeling.metamodel import MetaModel
from codelogician.modeling.model import Model
from codelogician.modeling.task_creator import ModelTaskCreator
from codelogician.util import filter, fst, maybe, maybe_else, not_none, translate

from ..common import Border, MyHeader, bind, decomp_ui, opaques_rich, vg_ui
from ..tree_views import TreeViews
from .model_cmds import ModelCommandsView


def code_view(title, lang, code, collapsed=True, max_height=16):
    from rich.syntax import Syntax

    with Collapsible(title=title, collapsed=collapsed) as c:
        c.styles.height = 'auto'
        with ScrollableContainer() as c:
            c.styles.height = 'auto'
            c.styles.max_height = max_height
            if 'pytest' in sys.modules:
                renderable = Syntax(
                    code, lang, line_numbers=True, indent_guides=True, theme='ansi_dark'
                )
            else:
                renderable = Syntax(code, lang, line_numbers=True, indent_guides=True)
            s = Static(renderable, id=lang, classes='code')
            s.styles.width = 'auto'
            s.styles.min_width = '100%'
            yield s


def view_opaques(opaques):
    header, table = opaques_rich(opaques)
    with Collapsible(title=header):
        yield Static(table)


def view_vgs(vgs):
    with Collapsible(title='Verification goals'):
        for i, vg in enumerate(vgs):
            yield Border('VG #%d' % (i + 1), vg_ui(vg))


def view_decomps(decomps):
    with Collapsible(title='Region decomposition'):
        for i, d in enumerate(decomps):
            yield Border('#%d' % (i + 1), decomp_ui(d))


def view_errors(errors):
    from regex import compile

    parse_kind = compile(r'\s*{\s*Kind.name\s*=\s*"(.*)"\s*}\s*').match
    kinds = {
        'SyntaxErr': 'Syntax error',
        'TypeErr': 'Type error',
        'TacticEvalErr': 'Tactic evaluation error',
    }

    def fmt_pos(pos):
        return f'line {pos.line}, col {pos.col}'

    def fmt_kind(kind):
        return translate(
            kinds, maybe_else(kind, lambda m: m.group(1), parse_kind(kind))
        )

    def error_ui(e, contexts):
        with VerticalGroup():
            label, msg = (
                Label(f'[bold]{fmt_kind(e.kind)}[/bold]: '),
                Static(f'[red]{e.msg.msg}[/]'),
            )
            yield from (
                [label, msg] if '\n' in e.msg.msg else [HorizontalGroup(label, msg)]
            )

            for loc, c in zip(e.msg.locs, contexts):
                start_l, start_c = c.start_line, c.start_col
                stop_l, stop_c = c.stop_line, c.stop_col
                file = c.file.replace('.py', '.iml')  # TODO: prefer not to hack this
                yield Static(f'in {file} {start_l}:{start_c} -- {stop_l}:{stop_c}')
                yield Static('\n'.join(c.lines))

    with Collapsible(title='Formalization errors'):
        for i, (e, contexts) in enumerate(errors):
            with Border('#%d' % (i + 1)):
                yield from error_ui(e, contexts)


def if_nonempty(f, xs):
    if len(xs) > 0:
        yield from f(xs)


class ModelStateView(VerticalGroup):
    DEFAULT_CSS = """StepView { margin: 0 1 0 0; }"""

    def __init__(self, model, container):
        VerticalGroup.__init__(self)
        self.model = model
        self.container = container

    def compose(self):
        fstate = self.model.agent_state
        if fstate is None:
            yield Static('Not formalised')
        else:
            errors = list(zip(fstate.errors, fstate.error_contexts))
            yield from code_view('IML code', 'ocaml', fstate.iml_code, collapsed=False)
            yield from if_nonempty(view_opaques, fstate.opaque_funcs)
            yield from if_nonempty(view_vgs, fstate.vgs)
            yield from if_nonempty(view_decomps, fstate.region_decomps)
            yield from if_nonempty(view_errors, errors)


class ModelView(VerticalGroup):
    DEFAULT_CSS = '#status { margin-top: 1; margin-bottom: 1 }'

    def __init__(self, model: Model, container):
        VerticalGroup.__init__(self)
        self.model = model
        self.container = container

    def compose(self):
        task_creator = ModelTaskCreator()

        yield Label('[$accent][b]%s[/b][/]' % self.model.rel_path)
        with VerticalScroll():
            yield Static(task_creator.gen_model_status(self.model), id='status')
            if self.model.src_code:
                yield from code_view(
                    'Src code', 'python', self.model.src_code, collapsed=False
                )
            yield ModelStateView(self.model, self.container)


class DecompsView(VerticalGroup):
    def __init__(self, state_of_path, paths: list[Path], container, *args, **kwargs):
        VerticalGroup.__init__(self, *args, **kwargs)
        self.state_of_path = state_of_path
        self.paths = paths

    def compose(self):
        with VerticalScroll():
            for p in self.paths:
                yield Label('[$primary][b]%s[/b][/]' % p)
                with VerticalGroup() as v:
                    state = self.state_of_path(p)
                    v.styles.padding = (0, 0, 0, 2)
                    if state is not None:
                        for i, decomp in enumerate(state.region_decomps):
                            yield Border('Decomp #%d' % (i + 1), decomp_ui(decomp))
                    else:
                        yield Static('[$foreground-muted]Not formalised[/]')
                yield Static('')


class VGsView(VerticalGroup):
    def __init__(self, state_of_path, paths: list[Path], container, *args, **kwargs):
        VerticalGroup.__init__(self, *args, **kwargs)
        self.state_of_path = state_of_path
        self.paths = paths

    def compose(self):
        with VerticalScroll():
            for p in self.paths:
                yield Label('[$primary][b]%s[/b][/]' % p)
                with VerticalGroup() as v:
                    state = self.state_of_path(p)
                    v.styles.padding = (0, 0, 0, 2)
                    if state is not None:
                        for i, vg in enumerate(state.vgs):
                            yield Border('VG #%d' % (i + 1), vg_ui(vg))
                    else:
                        yield Static('[$foreground-muted]Not formalised[/]')
                yield Static('')


class OpaquesView(VerticalGroup):
    def __init__(self, state_of_path, paths: list[Path], container, *args, **kwargs):
        VerticalGroup.__init__(self, *args, **kwargs)
        self.state_of_path = state_of_path
        self.paths = paths

    def compose(self):
        with VerticalScroll():
            for p in self.paths:
                yield Label('[$primary][b]%s[/b][/]' % p)
                with VerticalGroup() as v:
                    v.styles.padding = (0, 0, 0, 1)
                    state = self.state_of_path(p)
                    if state is not None:
                        _, table = opaques_rich(state.opaque_funcs)
                        yield Static(table)
                    else:
                        yield Static(' [$foreground-muted]Not formalised[/]')


class ModelScreen(Screen):
    DEFAULT_CSS = """
        #panes { width: 65% }
        #panes TabPane { padding: 0 1 0 1; }
        Label.subtree-descriptor { color: $accent; margin-bottom: 1; width: 1fr }
        OpaquesView { background: $surface }
        VGsView { background: $surface }
        """

    mmodel: reactive[None | MetaModel] = reactive(None)
    last_update: reactive[str] = reactive('-never-')
    selected: reactive[None | str] = reactive(None)

    def on_mount(self):
        self.title = 'Model View'
        bind(self, 'last_update', 'mmodel')

    def empty_container(self, id):
        container = self.query_one(id)
        container.remove_children()
        return container

    def watch_mmodel(self, old_mm: MetaModel, new_mm: MetaModel):
        if self.selected and self.selected in new_mm.models:
            self.select_path(self.selected)
        else:
            for id in ['#model_tab', '#opaques_tab', '#decomps_tab', '#vgs_tab']:
                self.empty_container(id).mount(Static('<Nothing selected>'))

    def compose(self):
        yield MyHeader()
        with Horizontal():
            yield TreeViews().data_bind(ModelScreen.mmodel)
            with TabbedContent(id='panes'):
                with TabPane('Model state', id='model_tab'):
                    yield Static('<Nothing selected>')
                with TabPane('Command entry', id='commands_tab'):
                    yield Static('<Nothing selected>')
                with TabPane('Opaque functions', id='opaques_tab'):
                    yield Static('<Nothing selected>')
                with TabPane('Decomposition requests', id='decomps_tab'):
                    yield Static('<Nothing selected>')
                with TabPane('Verification goals', id='vgs_tab'):
                    yield Static('<Nothing selected>')

        yield Footer()

    def watch_selected(self, _, selected):
        maybe(self.select_path, selected)

    def select_path(self, rel_path):
        # fmt: off
        tree = self.query_one(TreeViews).active_tree()
        file_path = None if rel_path == '' else tree.path_of_node(rel_path)
        children = tree.get_descendants(rel_path)
        paths = filter(not_none, [file_path, *map(tree.path_of_node, children)])
        assert tree.id is not None
        descriptor = ('All modules' if rel_path == '' else
                      { 'src-tree':   (f'Just the module [b]{rel_path}[/b]' if file_path else
                                       f'Modules in directory [b]{rel_path}[/b]:')
                      , 'rev-deps':    f'Modules [b]{rel_path}[/b] and its dependents:'
                      , 'module-deps': f'Module [b]{rel_path}[/b] and its dependencies:'
                      }[tree.id])

        def state_of_path(p):
            return cast(MetaModel, self.mmodel).models[p].agent_state

        def update_(id, mk_view):
            container = self.empty_container(id)
            container.mount(Label(descriptor, classes='subtree-descriptor'))
            container.mount(mk_view(state_of_path, paths, self))
        # fmt: on

        model = cast(MetaModel, self.mmodel).models.get(rel_path)
        if model:  #  and model.src_code:
            container = self.empty_container('#model_tab')
            container.mount(ModelView(model, self))

            container = self.empty_container('#commands_tab')
            container.mount(ModelCommandsView(model, self))

        update_('#opaques_tab', OpaquesView)
        update_('#decomps_tab', DecompsView)
        update_('#vgs_tab', VGsView)

    def on_tree_node_selected(self, event):
        self.selected = maybe_else('', fst, event.node.data)

    @on(Button.Pressed, '#send_commands')
    def add_model_commands(self, event):
        """Add commands to the model"""
        pass



#
#   Imandra Inc.
#
#   decomps.py
#

from textual.containers import (
    HorizontalGroup,
    VerticalGroup,
    VerticalScroll,
)
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Button, Footer, Label, Rule

from ..common import MyHeader, bind, decomp_ui


class DecompsScreen(Screen):
    """ """

    mmodel = reactive(None, recompose=True)

    def on_mount(self):
        self.title = 'Region Decompositions'
        bind(self, 'mmodel')

    # def watch_mmodel(self, old_value: MetaModel, new_value: MetaModel):
    #     """ """
    #     pass

    def compose(self):
        """ """

        yield MyHeader()
        with VerticalScroll():
            if self.mmodel:
                for idx, (path, model) in enumerate(self.mmodel.models.items()):
                    yield Rule()
                    yield Label(f'[$primary][b]{path}[/b][/]')
                    with HorizontalGroup():
                        yield Button('View model', id=f'btn_{idx}_view_model')
                        with VerticalGroup() as v:
                            v.styles.padding = (0, 0, 0, 1)
                            for decomp in model.decomps():
                                yield Rule()
                                yield decomp_ui(decomp)

        yield Footer()


if __name__ == '__main__':
    pass



#
# Imandra Inc.
#
# endpts_metamodel.py
#

import logging

from fastapi import HTTPException

from codelogician.modeling.mlist import ModelList, gen_listing

from ..cl_server import CLServer

log = logging.getLogger(__name__)


def register_metamodel_endpoints(app: CLServer):  # noqa: C901
    @app.get('/metamodel', operation_id='metamodel')
    async def latest_state():
        """
        Return the latest metaModel
        """
        log.info('Received request for the latest metamodel')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        # if state.curr_meta_model is None: return {}
        return state.curr_meta_model.toJSON()

    # Return summary of the current state
    @app.get('/metamodel/summary', operation_id='metamodel_summary')
    async def metamodel_latest_summary():
        """
        Return summary of the latest metamodel
        """

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        log.info('Received /state/summary request')
        # already returns JSON representation
        # if state.curr_meta_model is None:
        #    return {}
        # else:
        return state.curr_meta_model.summary()

    @app.get('/metamodel/list', operation_id='list_of_models')
    async def model_list(listby: str = 'alphabet'):
        """
        Return a list of model statistics, sorted by specified criteria. Options are:
        - None - then straight up list of models
        - frm_status - formalization status
        - upstream - number of models upstream which are affected by the specified model
        - opaques - number of opaque functions
        - failed_vgs - number of failed verification goals
        """

        log.info(f'Received request for model listing by {listby}.')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=400, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f'Failed to get strategy state: {e}'
            )

        try:
            if listby == 'alphabet':
                return ModelList(models=state.curr_meta_model.list_models())
            else:
                return gen_listing(models=state.curr_meta_model.models, category=listby)
        except Exception as e:
            raise HTTPException(
                status_code=400, detail=f'Error when requesting listing: {str(e)}'
            )

    @app.get('/metamodel/vgs', operation_id='list_of_vgs')
    async def verification_goals():
        """
        Return verification goals with their statuses for the entire metamodel
        """

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        log.info('Received /state/latest/vgs request')
        if state.curr_meta_model is None:
            return []

        return state.curr_meta_model.vgs()

    @app.get('/metamodel/decomps', operation_id='list_of_decomps')
    async def decomps():
        """
        Return the list of decomps
        """

        log.info('Received /metamodel/decomps request')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        if state.curr_meta_model is None:
            return []
        return state.curr_meta_model.decomps()

    @app.get('/metamodel/opaques', operation_id='list_of_opaque_functions')
    async def opaques():
        """
        Return list of opaque functions
        """

        log.info('Received /state/latest/opaques request')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        if state.curr_meta_model is None:
            return []
        return state.curr_meta_model.opaques()

    @app.get('/metamodel/cache/by_index/{idx}', operation_id='get_metamodel_from_cache')
    async def get_state(idx: int):
        """
        Returns a complete state of the cache.
        """
        log.info('Received cache metamodel request for index {idx}')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        if idx in state.meta_cache.indices():
            mmodel = state.meta_cache.get_cache_metamodel(idx)

            if mmodel is None:
                raise HTTPException(
                    status_code=404,
                    detail=f'Metamodel from cache with index={idx} not found!',
                )
            else:
                return mmodel.toJSON()
        else:
            log.error(f'{idx} index not found!')
            raise HTTPException(
                status_code=404,
                detail=f'Metamodel from cache with index={idx} not found!',
            )

    @app.get('/metamodel/cache/indices', operation_id='get_metamodel_cache_indices')
    async def cache_indices() -> list[int]:
        """
        Return the list of all indices in cache.
        """

        log.info('Request for all indices.')

        strategy_id = app.current_strat_id
        if strategy_id is None:
            raise HTTPException(status_code=404, detail='No current strategy')

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(
                status_code=404, detail=f'Failed to get strategy state: {e}'
            )

        return state.meta_cache.indices()



from pydantic import BaseModel, Field


class Position(BaseModel):
    line: int
    col: int


class Location(BaseModel):
    file: str | None = Field(default=None)
    start: Position | None = Field(default=None)
    stop: Position | None = Field(default=None)



from .artmsg import Art, StorageEntry
from .error import Error, ErrorKind, ErrorMessage
from .locs import Location, Position
from .session import Session, SessionCreate, SessionOpen
from .simple_api import (
    CounterSat,
    DecomposeReq,
    DecomposeRes,
    EvalOutput,
    EvalRes,
    EvalSrcReq,
    InferredType,
    InstanceNameReq,
    InstanceRes,
    InstanceSrcReq,
    LiftBool,
    Model,
    ModelType,
    OneshotReq,
    OneshotRes,
    PO_Res,
    Proved,
    Refuted,
    RegionStr,
    Sat,
    SessionCreateReq,
    TypecheckReq,
    TypecheckRes,
    Unsat,
    Verified_upto,
    VerifyNameReq,
    VerifyRes,
    VerifySrcReq,
)
from .task import Origin, Task, TaskID, TaskKind
from .utils import Empty, StringMsg

__all__ = [
    'Art',
    'CounterSat',
    'DecomposeReq',
    'DecomposeRes',
    'Empty',
    'Error',
    'ErrorKind',
    'ErrorMessage',
    'EvalOutput',
    'EvalRes',
    'EvalSrcReq',
    'InferredType',
    'InstanceNameReq',
    'InstanceRes',
    'InstanceSrcReq',
    'LiftBool',
    'Location',
    'Model',
    'ModelType',
    'OneshotReq',
    'OneshotRes',
    'Origin',
    'PO_Res',
    'Position',
    'Proved',
    'Refuted',
    'RegionStr',
    'Sat',
    'Session',
    'SessionCreate',
    'SessionCreateReq',
    'SessionOpen',
    'StorageEntry',
    'StringMsg',
    'Task',
    'TaskID',
    'TaskKind',
    'TypecheckReq',
    'TypecheckRes',
    'Unsat',
    'Verified_upto',
    'VerifyNameReq',
    'VerifyRes',
    'VerifySrcReq',
]



#
#   Imandra Inc.
#
#   codelogician/commands/cmd_eval.py
#

# pyright: strict
# Dev-doc

import asyncio
import json as jsonlib
import os
import sys
from pathlib import Path
from typing import Annotated, Any, Literal, TypedDict, assert_never, cast

import typer
from imandrax_api_models import DecomposeRes, InstanceRes, VerifyRes
from imandrax_api_models.client import (
    ImandraXAsyncClient,
    get_imandrax_async_client,
    get_imandrax_client,
)
from imandrax_api_models.context_utils import (
    format_decomp_res,
    format_eval_res,
    format_vg_res,
    remove_art_and_task_fields,
)
from imandrax_api_models.logging_utils import configure_logging
from iml_query.processing import (
    extract_decomp_reqs,
    extract_instance_reqs,
    extract_verify_reqs,
)
from iml_query.processing.decomp import DecompReqArgs
from iml_query.tree_sitter_utils import get_parser
from pydantic import TypeAdapter

from codelogician.util import get_imandra_uni_key

app = typer.Typer(name='ImandraX')


DEBUG = os.environ.get('DEBUG', '0') == '1'
"""Env var to enable debug logging."""

if DEBUG:
    configure_logging('debug')
else:
    configure_logging('warning')


def _remove_fields(data: dict[str, Any]) -> dict[str, Any]:
    """Resursively look inside a dict for certain keys and remove it."""
    data = data.copy()
    remove_fields = ['artifact', 'task']
    for k in list(data.keys()):
        v = data[k]
        if k in remove_fields:
            data.pop(k)
        elif isinstance(v, dict):
            v = cast(dict[str, Any], v)
            data[k] = _remove_fields(v)
    return data


def _load_iml(path: str | None) -> str:
    """Read IML code from a file or stdin."""
    if (path is None) or (path == '-'):
        # Read from stdin if no path is provided
        iml = sys.stdin.read()
    else:
        if not Path(path).exists():
            raise typer.BadParameter(f'IML file {path} does not exist')
        iml = Path(path).read_text()
    return iml


def ifprintf(message: Any | None) -> None:
    """Similar to OCmal's `ifprintf` function.

    Used to intercept stdout printing when "--json" is set.
    """
    pass


@app.command(
    name='check',
    help='Evaluate an IML file without VG and decomp.',
)
def check(
    file: Annotated[
        str | None,
        typer.Argument(
            help='Path of the IML file to evaluate. Set to "-" to read from stdin.',
        ),
    ] = None,
    with_vgs: Annotated[
        bool,
        typer.Option(
            help='Whether to strip verify and instance requests before evaluating.',
        ),
    ] = False,
    with_decomps: Annotated[
        bool,
        typer.Option(
            help='Whether to decomp requests before evaluating.',
        ),
    ] = False,
    json: Annotated[
        bool,
        typer.Option(
            help='Whether to output the results in JSON format.',
        ),
    ] = False,
):
    iml = _load_iml(file)

    c = get_imandrax_client(get_imandra_uni_key())
    eval_res = c.eval_model(src=iml, with_vgs=with_vgs, with_decomps=with_decomps)
    if not json:
        typer.echo(format_eval_res(eval_res, iml))
    else:
        typer.echo(eval_res.model_dump_json(indent=2))


# VG
# ====================


class VGItem(TypedDict):
    kind: Literal['verify', 'instance']
    src: str
    start_point: tuple[int, int]
    end_point: tuple[int, int]


def _collect_vgs(iml: str) -> list[VGItem]:
    tree = get_parser().parse(iml.encode('utf-8'))
    iml, tree, verify_reqs, verify_req_ranges = extract_verify_reqs(iml, tree)
    iml, tree, instance_reqs, instance_req_ranges = extract_instance_reqs(iml, tree)

    # Collect
    vg_items: list[VGItem] = []
    for req, req_range in zip(verify_reqs, verify_req_ranges, strict=True):
        vg_items.append(
            {
                'kind': 'verify',
                'src': req['src'],
                'start_point': (req_range.start_point[0], req_range.start_point[1]),
                'end_point': (req_range.end_point[0], req_range.end_point[1]),
            }
        )
    for req, req_range in zip(instance_reqs, instance_req_ranges, strict=False):
        vg_items.append(
            {
                'kind': 'instance',
                'src': req['src'],
                'start_point': (req_range.start_point[0], req_range.start_point[1]),
                'end_point': (req_range.end_point[0], req_range.end_point[1]),
            }
        )
    vg_items.sort(key=lambda x: x['start_point'])
    return vg_items


@app.command(
    name='list-vg',
    help='List verification goals in an IML file.',
)
def list_vg(
    file: Annotated[
        str | None,
        typer.Argument(
            help='Path of the IML file to check. Set to "-" to read from stdin.',
        ),
    ] = None,
    json: Annotated[
        bool,
        typer.Option(
            help='Whether to output the results in JSON format.',
        ),
    ] = False,
):
    iml = _load_iml(file)

    vgs: list[VGItem] = _collect_vgs(iml)

    if not json:
        for i, item in enumerate(vgs, 1):
            loc_str = (
                f'{item["start_point"][0]}:{item["start_point"][1]}'
                f'-{item["end_point"][0]}:{item["end_point"][1]}'
            )
            typer.echo(f'{i}: {item["kind"]} ({loc_str}): {item["src"]}')
    else:
        json_s = TypeAdapter(list[VGItem]).dump_json(vgs, indent=2)
        typer.echo(json_s)


@app.command(
    name='check-vg',
    help='Check verification goals in an IML file.',
)
def check_vg(
    file: Annotated[
        str | None,
        typer.Argument(
            help='Path of the IML file to check. Set to "-" to read from stdin.',
        ),
    ] = None,
    index: Annotated[
        list[int] | None,
        typer.Option(
            help='Name of the verification goal to check.',
        ),
    ] = None,
    check_all: Annotated[
        bool,
        typer.Option(
            help='Whether to check all verify requests in the IML file.',
        ),
    ] = False,
    json: Annotated[
        bool,
        typer.Option(
            help='Whether to output the results in JSON format.',
        ),
    ] = False,
):
    index = index or []

    if not json:
        echo = typer.echo
    else:
        echo = ifprintf

    async def _async_check_vg() -> list[VerifyRes | InstanceRes]:
        iml = _load_iml(file)
        vgs = _collect_vgs(iml)

        index_: list[int] = (
            list(range(1, len(vgs) + 1)) if check_all or (len(index) == 0) else index
        )

        vg_with_idx: list[tuple[int, VGItem]] = [
            (i, vg) for (i, vg) in enumerate(vgs, 1) if i in index_
        ]

        async def _check_vg(
            vg: VGItem,
            i: int,
            c: ImandraXAsyncClient,
        ) -> VerifyRes | InstanceRes:
            match vg['kind']:
                case 'verify':
                    res = await c.verify_src(src=vg['src'])
                case 'instance':
                    res = await c.instance_src(src=vg['src'])
                case _:
                    assert_never(vg['kind'])
            echo(f'{i}: {vg["kind"]} ({vg["src"]})')
            echo(format_vg_res(res))
            return res

        async with get_imandrax_async_client(get_imandra_uni_key()) as c:
            eval_res = await c.eval_model(src=iml)
            echo(format_eval_res(eval_res, iml))
            if eval_res.has_errors:
                echo('Error(s) found in IML file. Exiting.')
                sys.exit(1)
                return
            echo('\n' + '=' * 5 + 'VG' + '=' * 5 + '\n')
            tasks = [_check_vg(vg, i, c) for (i, vg) in vg_with_idx]
            return await asyncio.gather(*tasks)

    vg_res_list: list[VerifyRes | InstanceRes] = asyncio.run(_async_check_vg())
    if json:
        out: list[dict[str, Any]] = [
            remove_art_and_task_fields(vg_res.model_dump(mode='json'))
            for vg_res in vg_res_list
        ]
        typer.echo(jsonlib.dumps(out, indent=2))


# decomp
# ====================


class DecompItem(TypedDict):
    req_args: DecompReqArgs
    start_point: tuple[int, int]
    end_point: tuple[int, int]


def _collect_decomps(iml: str) -> list[DecompItem]:
    tree = get_parser().parse(iml.encode('utf-8'))
    iml, tree, decomp_reqs, ranges = extract_decomp_reqs(iml, tree)

    decomp_items: list[DecompItem] = [
        DecompItem(
            req_args=req,
            start_point=range_.start_point,
            end_point=range_.end_point,
        )
        for req, range_ in zip(decomp_reqs, ranges, strict=True)
    ]

    decomp_items.sort(key=lambda x: x['start_point'])
    return decomp_items


@app.command(
    name='list-decomp',
    help='List decomp requests in an IML file.',
)
def list_decomp(
    file: Annotated[
        str | None,
        typer.Argument(
            help='Path of the IML file to check. Set to "-" to read from stdin.',
        ),
    ] = None,
    json: Annotated[
        bool,
        typer.Option(
            help='Whether to output the results in JSON format.',
        ),
    ] = False,
):
    iml = _load_iml(file)
    decomps: list[DecompItem] = _collect_decomps(iml)

    if not json:
        for i, item in enumerate(decomps, 1):
            typer.echo(f'{i}: {item["req_args"]["name"]}')
    else:
        json_s = TypeAdapter(list[DecompItem]).dump_json(decomps, indent=2)
        typer.echo(json_s)


@app.command(
    name='check-decomp',
    help='Check decomp requests in an IML file.',
)
def check_decomp(
    file: Annotated[
        str | None,
        typer.Argument(
            help='Path of the IML file to check. Set to "-" to read from stdin.',
        ),
    ] = None,
    index: Annotated[
        list[int] | None,
        typer.Option(
            help='Index of the decomposition request to check.',
        ),
    ] = None,
    check_all: Annotated[
        bool,
        typer.Option(
            help='Whether to check all decomp requests in the IML file.',
        ),
    ] = False,
    json: Annotated[
        bool,
        typer.Option(
            help='Whether to output the results in JSON format.',
        ),
    ] = False,
):
    index = index or []

    if not json:
        echo = typer.echo
    else:
        echo = ifprintf

    async def _async_check_decomp() -> list[DecomposeRes]:
        iml = _load_iml(file)
        decomps = _collect_decomps(iml)

        index_: list[int] = (
            list(range(1, len(decomps) + 1))
            if check_all or (len(index) == 0)
            else index
        )

        decomp_with_idx: list[tuple[int, DecompItem]] = [
            (i, decomp) for (i, decomp) in enumerate(decomps, 1) if i in index_
        ]

        async def _check_decomp(
            decomp: DecompItem, i: int, c: ImandraXAsyncClient
        ) -> DecomposeRes:
            res = await c.decompose(**decomp['req_args'])
            echo(f'{i}: decompose {decomp["req_args"]["name"]}')
            echo(format_decomp_res(res))
            return res

        async with get_imandrax_async_client(get_imandra_uni_key()) as c:
            eval_res = await c.eval_model(src=iml)
            echo(format_eval_res(eval_res, iml))
            if eval_res.has_errors:
                typer.echo('Error(s) found in IML file. Exiting.')
                sys.exit(1)
                return

            echo('\n' + '=' * 5 + 'Decomp' + '=' * 5 + '\n')
            tasks = [_check_decomp(decomp, i, c) for (i, decomp) in decomp_with_idx]
            return await asyncio.gather(*tasks)

    decomp_res_list: list[DecomposeRes] = asyncio.run(_async_check_decomp())

    if json:
        out: list[dict[str, Any]] = [
            remove_art_and_task_fields(decomp_res.model_dump(mode='json'))
            for decomp_res in decomp_res_list
        ]
        typer.echo(jsonlib.dumps(out, indent=2))


# Conditionally add gen-test command
try:
    if not os.environ.get('IMANDRAX_API_KEY'):
        os.environ['IMANDRAX_API_KEY'] = get_imandra_uni_key()
    from imandrax_codegen.__main__ import (  # pyright: ignore[reportMissingTypeStubs]
        gen_test_command,
    )

    app.command(name='gen-test')(gen_test_command)
except ValueError as e:
    if 'Only MacOS is supported' in str(e):
        pass
    else:
        raise e


if __name__ == '__main__':
    app()



#
#   Imandra Inc.
#
#   base.py
#

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from io import StringIO
from typing import Literal, Self

import joblib
import networkx as nx
import structlog
from rich.console import Console
from rich.tree import Tree

from ..dep_tools.utils import get_network_text, nx_graph_to_mermaid

logger = structlog.get_logger(__name__)


class NodeMixin:
    @property
    @abstractmethod
    def path_key(self) -> str:
        """Primary key for the node. Used for indexing."""
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass


def nx_graph_from_adjacency_list[Node: NodeMixin](
    nodes: list[Node],
    adjacency_list: dict[str, set[str]],
) -> nx.DiGraph:
    """
    Args:
        - nodes: list of nodes
        - adjacency_list: adjacency list of the graph
            - key: node key (path_key)
            - value: set of node keys that the key node depends on
    Returns:
        - DependencyGraph
    """
    # Check pkey uniqueness
    if len({node.path_key for node in nodes}) != len(nodes):
        raise ValueError('Nodes must have unique path keys')

    def get_node_by_key(key: str) -> Node:
        maybe_node = next((node for node in nodes if node.path_key == key), None)
        if maybe_node is None:
            raise ValueError(f'Node with key {key} not found')
        return maybe_node

    nx_graph = nx.DiGraph()
    for node in nodes:
        nx_graph.add_node(node)

    for src_key, tgt_keys in adjacency_list.items():
        src_node = get_node_by_key(src_key)
        for tgt_key in tgt_keys:
            tgt_node = get_node_by_key(tgt_key)
            nx_graph.add_edge(src_node, tgt_node)

    return nx_graph


class DepGraph[Node: NodeMixin]:
    """
    Dependency graph
    - Edge type: Singleton {}
    - Node and edge meaning:
        - 1. (file | dir, contains)
        - 2. (module | package, import)
        - 3. (symbol, use)
    """

    def __init__(self, nx_graph: nx.DiGraph):
        # Check singleton edge type
        for _src, _tgt, data in nx_graph.edges(data=True):
            if data != {}:
                raise ValueError(f'Edge type must be `{{}}`, but got `{data}`')
        self.nx_graph = nx_graph

    def get_node_by_path(self, path: str) -> Node | None:
        return next((node for node in self.nodes if node.path_key == path), None)

    @property
    def nodes(self) -> list[Node]:
        return list(self.nx_graph.nodes)

    @property
    def is_tree(self) -> bool:
        """Connected and acyclic.

        Treat an empty graph as a valid tree for our purposes (e.g., empty repo dir).
        """
        if self.nx_graph.number_of_nodes() == 0:
            return True
        return nx.is_tree(self.nx_graph)

    @property
    def is_acyclic(self) -> bool:
        return nx.is_directed_acyclic_graph(self.nx_graph)

    ##########
    # Factory
    ##########

    @classmethod
    def from_adjacency_list(
        cls,
        nodes: list[Node],
        adjacency_list: dict[str, set[str]],
    ) -> DepGraph[Node]:
        nx_graph = nx_graph_from_adjacency_list(nodes, adjacency_list)
        return cls(nx_graph)

    def get_adjacency_list(self) -> dict[str, set[str]]:
        adj = dict(self.nx_graph.adjacency())
        return {
            node.path_key: {tgt.path_key for tgt in tgts} for node, tgts in adj.items()
        }

    ##########
    # Node getters
    ##########
    def get_all_nodes(
        self,
        order: Literal['leaves_to_root', 'root_to_leaves'] | None = 'leaves_to_root',
    ) -> list[Node]:
        if order in ['root_to_leaves', 'leaves_to_root']:
            topo_sorted_nodes = list(nx.topological_sort(self.nx_graph))
            if order == 'root_to_leaves':
                return topo_sorted_nodes
            else:
                return topo_sorted_nodes[::-1]
        else:
            return list(self.nx_graph.nodes)

    def get_successors(self, node: Node) -> list[Node]:
        return list(self.nx_graph.successors(node))

    def get_predecessors(self, node: Node) -> list[Node]:
        return list(self.nx_graph.predecessors(node))

    def get_orphan_nodes(self) -> list[Node]:
        return [
            node
            for node in self.nx_graph
            if not list(self.nx_graph.successors(node))
            and not list(self.nx_graph.predecessors(node))
        ]

    def get_terminal_nodes(self) -> list[Node]:
        return [
            node for node in self.nx_graph if not list(self.nx_graph.successors(node))
        ]

    def get_initial_nodes(self) -> list[Node]:
        return [
            node for node in self.nx_graph if not list(self.nx_graph.predecessors(node))
        ]

    def get_root_node(self) -> Node:
        if not self.is_tree:
            raise ValueError('Graph is not a tree')
        return self.get_initial_nodes()[0]

    def get_descendants(
        self,
        node: Node,
        order: Literal['leaves_to_root', 'root_to_leaves'] | None = 'leaves_to_root',
    ) -> list[Node]:
        descendants = list(nx.descendants(self.nx_graph, node))
        return (
            [n for n in self.get_all_nodes(order=order) if n in descendants]
            if order in ['root_to_leaves', 'leaves_to_root']
            else descendants
        )

    def get_ancestors(
        self,
        node: Node,
        order: Literal['leaves_to_root', 'root_to_leaves'] | None = 'leaves_to_root',
    ) -> list[Node]:
        ancestors = list(nx.ancestors(self.nx_graph, node))
        if order in ['root_to_leaves', 'leaves_to_root']:
            if order == 'root_to_leaves':
                return ancestors
            else:
                return ancestors[::-1]
        else:
            return ancestors

    ##########
    # Subgraph
    ##########

    def filter_nodes(self, predicate: Callable[[Node], bool]) -> Self:
        """
        Create a new DependencyGraph by removing nodes that satisfy the predicate
        condition.
        """
        # Get nodes to keep (inverse of the predicate)
        nodes_to_keep = [node for node in self.nx_graph.nodes if not predicate(node)]

        # Create subgraph with only the nodes we want to keep
        filtered_nx_graph = self.nx_graph.subgraph(nodes_to_keep).copy()

        return self.__class__(filtered_nx_graph)  # pyright: ignore

    ##########
    # IO
    ##########

    def dump(self, filename: str):
        nodes = self.nodes
        adj_list = self.get_adjacency_list()
        data = {'nodes': nodes, 'adj_list': adj_list}

        return joblib.dump(data, filename)

    @classmethod
    def load(cls, filename: str):
        data = joblib.load(filename)
        nodes = data['nodes']
        adj_list = data['adj_list']
        return cls.from_adjacency_list(nodes, adj_list)

    ##########
    # Repr
    ##########

    def edges_repr(self) -> str:
        return '\n'.join(f'{src!r} --> {tgt!r}' for src, tgt in self.nx_graph.edges())

    def graph_repr(self, method: Literal['rich', 'nx'] = 'rich') -> str:
        if method == 'nx':
            return get_network_text(self, ascii_only=True)
        else:
            adj_list = self.get_adjacency_list()
            nodes = self.nodes

            def pkey_to_repr(pkey: str) -> str:
                node = next((node for node in nodes if node.path_key == pkey), None)
                if node is None:
                    raise ValueError(f'Node with path key {pkey} not found')
                return repr(node)

            adj_repr: dict[str, list[str]] = {
                pkey_to_repr(src): [pkey_to_repr(tgt) for tgt in tgts]
                for src, tgts in adj_list.items()
            }
            # sort by src repr
            adj_repr = dict(sorted(adj_repr.items()))
            # sort by tgt repr
            adj_repr = {src: sorted(tgts) for src, tgts in adj_repr.items()}

            trees = []
            for src, tgts in adj_repr.items():
                tree = Tree(src)
                for tgt in tgts:
                    tree.add(tgt)
                trees.append(tree)

            string_buffer = StringIO()
            console = Console(file=string_buffer, width=80, record=True)
            for tree in trees:
                console.print(tree)
            return console.export_text()

    def _repr_markdown_(self) -> str:
        return nx_graph_to_mermaid(self.nx_graph)



#
#   Imandra Inc.
#
#   api_ref.py
#

import json
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Final, Self
from uuid import UUID

import structlog
import typer

logger = structlog.get_logger(__name__)


@dataclass
class IMLAPIReference:
    id: UUID
    module: str
    name: str
    signature: str
    doc: str | None = None
    pattern: str | None = None

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> Self:
        return cls(**{k: v for k, v in d.items() if k in cls.__dataclass_fields__})

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)


_iar_cache: list[IMLAPIReference] | None = None

DOC_DIR: Final[Path] = Path(__file__).parent.parent / 'data'
SELECTED_GUIDE_DOC: Final[list[str]] = [
    'guides/tactics.md',
    'guides/ordinal.md',
    'guides/termination_proofs.md',
    'guides/region-decomp/quick-start.md',
    'guides/region-decomp/writing-decomp.md',
    'guides/region-decomp/how-to-use-decomp-for-analysis.md',
    'guides/region-decomp/examples/',
    'guides/verification-goal/quick-start.md',
    'guides/verification-goal/examples/',
]


@dataclass
class MDDoc:
    path: Path  # path relative to DOC_DIR
    title: str
    content: str


md_docs: list[MDDoc] | None = None


def _load_guide_docs() -> list[MDDoc]:
    """
    Load documents in guide/ directory.
    """
    global md_docs

    def get_doc(doc_rel_path: Path) -> MDDoc:
        content = (DOC_DIR / doc_rel_path).read_text()
        return MDDoc(
            doc_rel_path,
            doc_rel_path.stem,
            content,
        )

    if md_docs is None:
        md_docs = []
        for doc_path in SELECTED_GUIDE_DOC:
            abs_path = DOC_DIR / doc_path
            if abs_path.is_file():
                logger.debug(f'Loading guide doc file: {abs_path}')
                rel_path = abs_path.relative_to(DOC_DIR)
                md_docs.append(get_doc(rel_path))
            else:
                for doc_file in abs_path.glob('**/*.md'):
                    rel_path = doc_file.relative_to(DOC_DIR)
                    md_docs.append(get_doc(rel_path))
    return md_docs


def _load_iar() -> list[IMLAPIReference]:
    """
    Load IML API Reference data lazily.
    """
    global _iar_cache
    if _iar_cache is None:
        iar_path = DOC_DIR / 'iml_api_reference_202510011126.json'
        iar_data = json.loads(iar_path.read_text())
        _iar_cache = [IMLAPIReference.from_dict(i) for i in iar_data]
    return _iar_cache


def dump_iml_overview_and_api_refs(dir_path: Path):
    """
    Writes the IML overview and API reference to a directory.

    Output will be like this:
        - iml_overview.md
        - iml_api_reference/
            - global.md
            - module1.md
            - module2.md
            - etc...
    """
    if not dir_path.exists():
        dir_path.mkdir(exist_ok=False, parents=True)

    if not dir_path.is_dir():
        typer.secho(
            f'Path is not a directory: {dir_path}', fg=typer.colors.RED, err=True
        )
        raise typer.Exit(1)

    from codelogician.doc.utils.prompts import (
        iml_caveats,
        iml_intro,
        lang_agnostic_meta_eg_overview,
        lang_agnostic_meta_egs_str,
    )

    # Write overview
    iml_overview = (
        iml_intro
        + iml_caveats
        + lang_agnostic_meta_eg_overview
        + lang_agnostic_meta_egs_str
    )
    overview_path = dir_path / 'iml_overview.md'
    overview_path.write_text(iml_overview)
    typer.secho(f'✓ Written overview to {overview_path}', fg=typer.colors.GREEN)

    # Write API reference
    iar = _load_iar()
    api_ref_dir = dir_path / 'iml_api_reference'
    api_ref_dir.mkdir(exist_ok=True)

    # Group by module
    modules = sorted({entry.module for entry in iar})
    for module in modules:
        module_entries = [entry for entry in iar if entry.module == module]
        module_name = module or 'global'

        is_global: bool = module_name == 'global'

        # Create markdown content for this module
        module_lines = [
            f'# Module `{module_name}`\n' if not is_global else '# Global\n',
            f'\n{len(module_entries)} entries\n\n',
        ]

        if is_global:
            module_lines.append(
                f'The following entries are qualified by the module `{module_name}`.'
                '\n\n'
            )

        for entry in module_entries:
            module_lines.append(f'`{entry.name}`\n')
            module_lines.append(f'- Signature: `{entry.signature}`\n')
            if entry.doc:
                module_lines.append(f'- Doc: {entry.doc}.\n')
            # if entry.pattern:
            #     module_lines.append(f"\n**Pattern:** `{entry.pattern}`\n")
            module_lines.append('\n')

        # Write module file
        module_filename = f'{module_name}.md' if module else 'global.md'
        module_path = api_ref_dir / module_filename
        module_path.write_text(''.join(module_lines))

    typer.secho(
        f'✓ Written API reference to {api_ref_dir} ({len(modules)} modules)',
        fg=typer.colors.GREEN,
    )


def dump_guide_docs(dir_path: Path):
    """
    Writes the guide documents to a directory.

    - contract: The total number of dumped markdown files will always be >= len(SELECTED_GUIDE_DOC).
    This contract must be respected and is verified by tests.
    - deps: global SELECTED_GUIDE_DOC
    - effect: writes guide documents to dir_path

    Args:
        dir_path: The directory where guide documents will be written

    Output structure example:
        guides/
            tactics.md
            ordinal.md
            termination_proofs.md
            region-decomp/
                quick-start.md
                writing-decomp.md
                examples/
                    _index.md
                    basics/_index.md
                    ...
    """
    if not dir_path.exists():
        dir_path.mkdir(exist_ok=False, parents=True)

    if not dir_path.is_dir():
        typer.secho(
            f'Path is not a directory: {dir_path}', fg=typer.colors.RED, err=True
        )
        raise typer.Exit(1)

    guide_docs = _load_guide_docs()
    for doc in guide_docs:
        out_path = dir_path / doc.path
        out_path.parent.mkdir(exist_ok=True, parents=True)
        out_path.write_text(doc.content)
        typer.secho(f'✓ Written {doc.title} to {out_path}', fg=typer.colors.GREEN)



from typing import NamedTuple, cast

from imandra.u.agents.code_logician.base import FormalizationState
from imandrax_api_models import ErrorMessage, Position

from ..util import filter, find, flat_map, map, scanl, snd


class ErrorSnippet(NamedTuple):
    file: str
    start_line: int
    start_col: int
    stop_line: int
    stop_col: int
    lines: list[str]


def full_iml_model(fstate: FormalizationState):
    """Get the full IML code, including the monolith IML redefining the dependencies
    modules and the original IML code.
    """
    if fstate.iml_model is None:
        raise ValueError('IML code is required')

    def fmt_module(module):
        indented = ['    ' + l for l in module.content.strip().splitlines()]
        lines = [f'module {module.name} = struct', *indented, 'end', '']
        return str(module.relative_path), lines

    main = None, fstate.iml_model.splitlines()
    return [fmt_module(d.iml_module) for d in fstate.dependency] + [main]


def join_sections(sections):
    return flat_map(snd, sections)


def find_line(sections, line):
    cum_sums = list(scanl(lambda x, y: x + len(snd(y)), 0, sections))
    n = cast(int, find(lambda i: line < cum_sums[i], range(len(cum_sums))))
    return line - cum_sums[n - 1], sections[n - 1]


def snippet_lines(
    lines: list[str], start: Position, end: Position, context_line: int = 2
):
    """Format a code snippet with highlighted error range."""
    start_line, start_col = start.line, start.col - 1
    end_line, end_col = end.line, end.col

    # Determine visible range
    visible_start = max(0, start_line - context_line - 1)
    visible_end = min(len(lines), end_line + context_line)

    # Calculate max line number width for alignment
    line_no_width = len(str(visible_end))
    underline_prefix = ' ' * (2 + line_no_width)

    def line(underline):
        return f'{underline_prefix} | {underline}'

    for i in range(visible_start, visible_end):
        line_no = i + 1

        # Mark lines within error range
        marker = '*' if start_line <= line_no <= end_line else ' '
        yield f'{marker} {line_no:{line_no_width}} | {lines[i]}'

        # Add underline/pointer
        if line_no == start_line == end_line:
            # Single line error
            yield line(' ' * start_col + '^' * (end_col - start_col))
        elif line_no == start_line:
            # Start of multi-line error
            yield line(' ' * start_col + '^' + '~' * (len(lines[i]) - start_col))
        elif line_no == end_line:
            # End of multi-line error
            yield line('~' * end_col)
        elif start_line < line_no < end_line:
            # Middle of multi-line error
            yield line('~' * len(lines[i]))


def describe_error(mod_path: str, sections, error_msg: ErrorMessage):
    def has_loc(loc):
        return loc.start is not None and loc.stop is not None

    def snippet(p) -> ErrorSnippet:
        n = p.stop.line - p.start.line
        l, (rel_path, lines) = find_line(sections, p.start.line - 1)
        ll, lines, path = (
            (l, lines[1:], rel_path) if rel_path else (l + 1, lines, mod_path)
        )
        start = Position(line=ll, col=p.start.col)
        stop = Position(line=ll + n, col=p.stop.col)
        snippet = list(snippet_lines(lines, start, stop))

        return ErrorSnippet(path, start.line, start.col, stop.line, stop.col, snippet)

    return map(snippet, filter(has_loc, error_msg.locs or []))



#
#   Imandra Inc.
#
#   sketches.py
#

from textual import on
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import (
    Button,
    Footer,
    Input,
    Label,
    ListItem,
    ListView,
    Rule,
    Select,
    Static,
    TabbedContent,
    TabPane,
    TextArea,
)

from codelogician.sketch.container import SketchContainer
from codelogician.sketch.sketch import (
    Sketch,
    SketchState,
)
from codelogician.ui.common import MyHeader, bind
from codelogician.util import maybe


class SketchScreen(Screen):
    """
    SketchScreen
    """

    #    mmodel = reactive("", recompose=True)
    sketch_container: reactive[None | SketchContainer] = reactive(None, recompose=True)
    selected_sketch: reactive[None | Sketch] = reactive(None)
    selected_sketch_state: reactive[None | SketchState] = reactive(None)

    def __init__(self):
        """ """
        Screen.__init__(self)

        self.title = 'Sketches'
        bind(self, 'sketch_container')

    def upd_static(self, selector, text):
        self.query_one(selector, Static).update(text)

    def watch_selected_sketch(self):
        """
        Updated state of the selected sketch
        """

        sk: None | Sketch = self.selected_sketch

        if not sk:
            return
        # fmt: off
        self.upd_static('#sk_sketch_id', f'ID : {sk.sketch_id}')
        self.upd_static('#sk_anchor_model_path', f'Anchor model: {sk.anchor_model_path}')
        self.upd_static('#sk_processing', f'Is processing: {sk.processing}')
        # fmt: on

        state_list: ListView = self.query_one('#sketch_state_list_view', ListView)
        state_list.clear()
        state_list.extend([ListItem(Static(str(idx))) for idx in sk.state_ids()])
        state_list.index = 0  # let's select the first one

        # Let's also make sure that the next tab has 'Existing sketch' selected
        self.query_one('#tabbed_content', TabbedContent).active = 'tab_existing'

    def watch_selected_sketch_state(self):
        """
        Update the specific sketch state
        """

        state: None | SketchState = self.selected_sketch_state

        if not state:
            return

        # fmt: off
        self.upd_static('#sketch_state_id', f'State id: {state.state_id}')
        self.upd_static('#sketch_state_change', f'State change: {str(state.change)}')
        self.upd_static('#sketch_state_frm_status', f'Formalization status: {str(state.status)}')
        self.upd_static('#sketch_state_error', f'Error: {state.error if state.error else "N/A"}')
        # fmt: on
        self.query_one('#sketch_state_iml', TextArea).text = state.iml_code

    def compose(self) -> ComposeResult:
        """ """
        yield MyHeader()

        with Horizontal() as h:
            h.styles.layout = 'grid'
            h.styles.grid_size_columns = 2
            h.styles.grid_columns = '1fr 3fr'

            with Vertical():
                yield Static('Existing sketches:')
                if self.sketch_container:
                    if self.sketch_container.ids():
                        items = []
                        for sketch_id in self.sketch_container.ids():
                            items.append(ListItem(Label(sketch_id)))
                        yield (ListView(*items, id='sketch_list_view'))
                    else:
                        yield Static('N/A')
                else:
                    yield Static('N/A')

            with TabbedContent(id='tabbed_content'):
                with TabPane('Existing', id='tab_existing'):
                    with Vertical():
                        yield Static('ID: N/A', id='sk_sketch_id')
                        yield Static('Anchor model: N/A', id='sk_anchor_model_path')
                        yield Static('Is processing:  N/A', id='sk_processing')
                        with Horizontal():
                            yield ListView(id='sketch_state_list_view')
                            yield Rule('vertical')
                            with Vertical():
                                yield Static('State id: N/A', id='sketch_state_id')
                                yield Static('State id: N/A', id='sketch_state_change')
                                yield Static(
                                    'Formalization status: N/A',
                                    id='sketch_state_frm_status',
                                )
                                yield Static('Error: N/A', id='sketch_state_error')
                                yield TextArea.code_editor(
                                    'N/A',
                                    language=None,
                                    id='sketch_state_iml',
                                    read_only=True,
                                )

                with TabPane('Create new', id='tab_new'):
                    with Vertical():
                        yield Static('Select anchor model:')
                        yield Select([])  # TODO: populate this
                        yield Input('Name')
                        yield Button('Create sketch', id='btn_create_sketch')

        yield Footer()

    @on(ListView.Highlighted, '#sketch_state_list_view')
    def on_sketch_state_highlighted(self):
        idx = self.query_one('#sketch_state_list_view', ListView).index
        if idx is not None and self.selected_sketch is not None:
            self.selected_sketch_state = self.selected_sketch.get_state_by_idx(idx)

    @on(ListView.Highlighted, '#sketch_list_view')
    def on_list_view_highlighted(self, event: ListView.Highlighted):
        """An existing sketch has been selected from this"""
        if self.sketch_container is not None:
            idx = self.query_one('#sketch_list_view', ListView).index
            sketches = list(self.sketch_container.sketches.values())
            self.selected_sketch = maybe(sketches.__getitem__, idx)

    @on(Button.Pressed, '#btn_create_sketch')
    def on_create_sketch(self):
        """ """
        pass


class TestSketchesApp(App):
    """
    Minimalist setup to test out Sketches screen
    """

    SCREENS = {'sketches': SketchScreen}

    def __init__(self, sk):
        super().__init__()
        self.sk = sk

    def compose(self):
        yield SketchScreen()

    def on_mount(self):
        self.push_screen('sketches')


if __name__ == '__main__':
    import json

    with open('src/codelogician/data/sketches/sketch1.json') as inFile:
        j = json.load(inFile)
        sketchC = SketchContainer.fromJSON(j)

    app = TestSketchesApp(sketchC)
    app.run()



from enum import Enum

from pydantic import BaseModel, Field


class TaskKind(Enum):
    TASK_UNSPECIFIED = 'TASK_UNSPECIFIED'
    TASK_EVAL = 'TASK_EVAL'
    TASK_CHECK_PO = 'TASK_CHECK_PO'
    TASK_PROOF_CHECK = 'TASK_PROOF_CHECK'
    TASK_DECOMP = 'TASK_DECOMP'


class TaskID(BaseModel):
    id: str = Field(description='The task identifier')


class Task(BaseModel):
    id: TaskID | None = Field(default=None)
    kind: TaskKind


class Origin(BaseModel):
    from_sym: str = Field(description='Symbol from which the task originated')
    count: int = Field(description='A counter for tasks for this symbol')



from __future__ import annotations

from enum import IntEnum
from pathlib import Path
from typing import ClassVar, Literal, Self

import networkx as nx
from pydantic import BaseModel, Field, model_validator

from ..tools.base import (
    DepGraph,
    NodeMixin,
    nx_graph_from_adjacency_list,
)

# from ..tools.filesystem import Dir, File, FileSystem


class SymbolKind(IntEnum):
    """A symbol kind.
    Imported from multilspy.lsp_protocol_handler.lsp_types.SymbolKind.
    """

    File = 1
    Module = 2
    Namespace = 3
    Package = 4
    Class = 5
    Method = 6
    Property = 7
    Field = 8
    Constructor = 9
    Enum = 10
    Interface = 11
    Function = 12
    Variable = 13
    Constant = 14
    String = 15
    Number = 16
    Boolean = 17
    Array = 18
    Object = 19
    Key = 20
    Null = 21
    EnumMember = 22
    Struct = 23
    Event = 24
    Operator = 25
    TypeParameter = 26


Range = tuple[tuple[int, int], tuple[int, int]]


class Symbol(BaseModel, NodeMixin):
    """A mix of LSP's SymbolInformation and DocumentSymbol."""

    name: str
    relative_path: Path = Field()
    kind: SymbolKind
    range: Range | None = Field(
        default=None,
        description='The range of the symbol. None represents a file-level symbol.',
    )
    detail: str | None = Field(default=None)
    container_name: str | None = Field(default=None, alias='containerName')
    children: list[Symbol] = Field(default_factory=list)

    @property
    def path_key(self) -> str:
        return str(self.relative_path)

    def __repr__(self) -> str:
        return str(self.relative_path)


class Module(Symbol):
    """A module."""

    name: str | None = Field(
        default=None,
        description='The name of the module. Automatically generated from the relative '
        'path for Python and IML modules.',
    )
    kind: Literal[SymbolKind.Module] = Field(default=SymbolKind.Module)
    content: str
    src_lang: str

    def __hash__(self) -> int:
        return hash((self.relative_path, self.content))

    @model_validator(mode='after')
    def validate_path(self) -> Self:
        if self.src_lang == 'python' and not self.relative_path.suffix == '.py':
            raise ValueError('Python module must have a .py extension')
        elif self.src_lang == 'iml' and not self.relative_path.suffix == '.iml':
            raise ValueError('IML module must have a .iml extension')
        return self


class ModuleDep[T: Module](DepGraph[T]):
    FILE_EXTENSION_MAP: ClassVar[dict[str, str]] = {
        'python': '.py',
        'iml': '.iml',
    }

    def __init__(self, nx_graph: nx.DiGraph):
        # Allow empty module graphs (e.g. repo has no matching source files)
        if nx_graph.number_of_nodes() == 0:
            # Choose a neutral default; callers must handle empty graphs anyway.
            self.src_lang = 'python'
            super().__init__(nx_graph)
            return

        # Check if all nodes have the same source language
        src_langs = {node.src_lang for node in nx_graph.nodes}
        if len(src_langs) != 1:
            raise ValueError(
                f'All nodes must have the same source language, but got {src_langs}'
            )

        # Check if the source language is supported
        src_lang = next(iter(src_langs))
        if src_lang not in self.FILE_EXTENSION_MAP:
            raise ValueError(
                f'Unsupported source language: {src_lang}. Supported languages: '
                f'{list(self.FILE_EXTENSION_MAP.keys())}'
            )

        self.src_lang = src_lang
        super().__init__(nx_graph)

    @classmethod
    def from_adjacency_list(
        cls,
        nodes: list[T],
        adjacency_list: dict[str, set[str]],
    ) -> Self:
        nx_graph = nx_graph_from_adjacency_list(
            nodes=nodes, adjacency_list=adjacency_list
        )
        return cls(nx_graph)



#
#   Imandra Inc.
#
#   processor.py
#

from __future__ import annotations

import logging
import os
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from threading import Timer

from ..modeling.dep_context import ModelDepContext
from ..modeling.task import ModelTask, UserManualIMLEditTask
from ..server.events import FileSystemCheckbackEvent, FileSystemEvent, ServerEvent
from ..sketch.task import SketchChangeTask
from .config import StratConfig, StratMode
from .events import (
    AutoModeCLTaskEvent,
    ChangeMode,
    CLResultEvent,
    ModelCLTaskEvent,
    RunOneshotEvent,
    SketchChangeEvent,
    SketchChangeResultEvent,
    StrategyEvent,
)
from .state import StrategyState
from .worker import CodeLogicianWorker

log = logging.getLogger(__name__)

EmitFn = Callable[[StrategyEvent | ServerEvent | None], None]


@dataclass
class PyIMLStrategyProcessor:
    """
    Testable event processor for PyIMLStrategy.

    - No Queue management here.
    - Emits follow-up events via `self.emit(...)`.
    """

    state: StrategyState
    config: StratConfig
    oneshot: bool = False

    _emit: EmitFn | None = None
    _worker: CodeLogicianWorker | None = None

    _language: str = 'Python'
    _extensions: list[str] = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        if self._extensions is None:
            self._extensions = ['.py', '.iml']

        if self.config is None:
            self.config = StratConfig()

        # clear any task IDs that might have been persisted
        self.state.clear_task_ids()

    # -------------------------
    # Wiring
    # -------------------------

    def set_emit(self, emit: EmitFn) -> None:
        self._emit = emit

    def emit(self, event: StrategyEvent | ServerEvent | None) -> None:
        if self._emit is None:
            raise RuntimeError('Processor emit callback is not set')
        self._emit(event)

    # -------------------------
    # Public helpers
    # -------------------------

    def language(self) -> str:
        return self._language

    def extensions(self) -> list[str]:
        return list(self._extensions)

    def update_config(self, new_config: StratConfig) -> None:
        self.config = new_config

    # -------------------------
    # Lifecycle
    # -------------------------

    def start(self) -> None:
        """
        Initialize worker, do initial sync if needed.
        """
        if self._worker is None:
            self._worker = CodeLogicianWorker(self.emit)
            self._worker.start()

        if self.config.mode == StratMode.AUTO:
            log.info('Starting file sync')
            self.state.run_file_sync()
            self._check_autoformalization_tasks()

        log.info('Started CodeLogician Worker thread')

    def shutdown(self) -> None:
        """
        Shut down worker (by sending None task).
        """
        if self._worker is not None:
            self._worker.add_task(None)

    # -------------------------
    # Event dispatch
    # -------------------------

    def process_event(self, event: ServerEvent | StrategyEvent) -> None:
        if isinstance(event, FileSystemEvent):
            log.info('About to process file system update event')
            self._proc_filesystem_update(event)
            return

        if isinstance(event, FileSystemCheckbackEvent):
            log.info('About to process file system checkback event')
            self._proc_filesystem_checkback(event)
            return

        if isinstance(event, ChangeMode):
            log.info(f'About to process mode change to: {str(event)}')
            self._proc_mode_change(event.new_mode)
            return

        if isinstance(event, CLResultEvent):
            log.info('About to process CL Result event')
            self._on_cl_result(event)
            return

        if isinstance(event, RunOneshotEvent):
            log.info('About to process RunOneshotEvent')
            self._proc_oneshot_event(event)
            return

        if isinstance(event, ModelCLTaskEvent):
            log.info('About to process ModelCLTaskEvent')
            self._proc_model_cl_task_event(event)
            return

        if isinstance(event, AutoModeCLTaskEvent):
            log.info('About to process AutoModeCLTaskEvent')
            self._proc_auto_mode_cl_task_event(event)
            return

        if isinstance(event, SketchChangeEvent):
            self._proc_sketch_change_event(event)
            return

        if isinstance(event, SketchChangeResultEvent):
            self._proc_sketch_change_result(event)
            return

        log.error(f'Gotten an unknown event type: {type(event).__name__}')

    # -------------------------
    # Sketch handling
    # -------------------------

    def _proc_sketch_change_event(self, event: SketchChangeEvent) -> None:
        if self._worker is None:
            raise RuntimeError('Worker not started')
        self._worker.add_task(
            SketchChangeTask(
                sketch_id=event.sketch_id, change=event.change, iml_code=''
            )
        )

    def _proc_sketch_change_result(self, event: SketchChangeResultEvent) -> None:
        log.info(f'Received ImandraX result for sketch={event.sketch_id}')
        sketch = self.state.sketches.from_id(event.sketch_id)
        if sketch:
            sketch.apply_change_result(event.change_result)
            log.info('Successfully applied it now.')
        else:
            log.error(f'Unknown sketch with id={event.sketch_id}')

    # -------------------------
    # Auto mode tasks
    # -------------------------

    def _proc_auto_mode_cl_task_event(self, event: AutoModeCLTaskEvent) -> None:
        if self._worker is None:
            raise RuntimeError('Worker not started')
        self._worker.add_task(event.task)
        log.info('Adding new task')
        self.state.add_task(event.task)

    def _proc_model_cl_task_event(self, event: ModelCLTaskEvent) -> None:
        mmodel = self.state.curr_meta_model
        if not mmodel or event.rel_path not in mmodel.models:
            log.error(
                "Somethings wrong - either the MetaModel doesn't exist or the model is missing "
                f'[{event.rel_path}]'
            )
            return

        oldGraphState = None
        model = mmodel.models[event.rel_path]
        if model.agent_state:
            oldGraphState = model.agent_state.graph_state

        clTask = ModelTask(
            rel_path=event.rel_path,
            specified_commands=[event.cmd],
            graph_state=oldGraphState,
        )

        log.info('Adding new task')
        self.state.add_task(clTask)

    def _proc_oneshot_event(self, _event: RunOneshotEvent) -> None:
        self.oneshot = True
        self.config.mode = StratMode.AUTO

        log.info(
            "We're now in AUTO mode - will check what we should do with autoformalization..."
        )
        self._check_autoformalization_tasks()

    def _proc_mode_change(self, newMode: StratMode) -> None:
        self.config.mode = newMode
        log.info(f'Setting strategy mode to {newMode}')

        if newMode == StratMode.AUTO:
            log.info("Since we're in AUTO mode now, will check formalization tasks")
            self._check_autoformalization_tasks()

    def _check_autoformalization_tasks(self) -> None:
        log.info('Checking autoformalization tasks!')

        tasks = self.state.get_next_tasks()
        if tasks:
            log.info(f'Adding {len(tasks)} new tasks')
            for task in tasks:
                self.emit(AutoModeCLTaskEvent(task=task))
            return

        log.info('No new formalization tasks present!')

        # oneshot: exit when no pending tasks
        if self.oneshot and self.state.tasks == []:
            log.info('There are no tasks now. Quitting!')
            self.emit(None)

    # -------------------------
    # File system event handling
    # -------------------------

    def _check_back_callback(self, abs_path: str) -> None:
        log.info(f'Adding FileSystemCheckbackEvent for {abs_path}')
        self.emit(FileSystemCheckbackEvent(abs_path=abs_path))

    def _proc_filesystem_update(self, event: FileSystemEvent) -> None:
        # Apply the immediate point update (src/iml content, create/delete/move, etc.)
        self.state.curr_meta_model.process_filesystem_event(event)

        if self.config.mode != StratMode.AUTO:
            log.info("No checkback thread started as we're in STANDBY mode")
            return

        # Only schedule checkback for file types we actually use for tasks/deps.
        ext = Path(event.abs_path1).suffix
        if ext not in ('.py', '.iml'):
            log.info(f'Skipping checkback for unsupported extension: {ext}')
            return

        log.info(
            "Since we're in AUTO mode, scheduling a Timer to check back and "
            'see if the change is stable enough to formalize.'
        )

        timer = Timer(
            interval=self.config.time_for_filesync_checkback,
            function=self._check_back_callback,
            args=(event.abs_path1,),
        )
        timer.start()

    def _proc_filesystem_checkback(self, event: FileSystemCheckbackEvent) -> None:
        log.info('Processing FileSystemCheckbackEvent')

        if self.state.curr_meta_model is None:
            log.info('No current MetaModel. Aborting.')
            return

        ext = Path(event.abs_path).suffix

        if ext == '.py':
            if not self.state.curr_meta_model.abs_path_in_models(event.abs_path):
                log.info(f'Model for path: {event.abs_path} not found. Aborting.')
                return

        elif ext == '.iml':
            model_path = str(Path(event.abs_path).with_suffix('.py'))
            if not self.state.curr_meta_model.abs_path_in_models(model_path):
                log.info(
                    f'IML code for model with path: {model_path} not found. Aborting.'
                )
                return
        else:
            return

        # Always resync models & dependencies once the filesystem has settled
        # Note: here we'll also check for cycles in dependency graphs
        log.info('Running MetaModel filesystem sync after checkback')
        try:
            self.state.curr_meta_model.run_file_sync()
        except Exception as e:
            log.error(f'Failed to run MetaModel file sync: {e}')
            return

        if self.config.mode == StratMode.AUTO:
            self._check_autoformalization_tasks()

    def _write_model_code(self, rel_path: str) -> None:
        mmodel = self.state.curr_meta_model
        if not mmodel:
            log.error(
                "Somehow we're processing a CLResult event but we have no current MetaModel!"
            )
            return

        models = mmodel.models
        if rel_path not in models:
            log.info('Somehow the result is for a non-existent model')
            return

        model = models[rel_path]

        if not model.iml_code():
            return

        if self.config.write_models_in_same_dir:
            file_name = Path(rel_path).with_suffix('.iml')
            tgt_path = os.path.join(self.state.src_dir_abs_path, file_name)

            log.info(f'Will now write out IML model to: {tgt_path}')
            try:
                with open(tgt_path, 'w') as outfile:
                    import_statements = ModelDepContext.create_import_iml_statements(
                        model
                    )
                    full_model = f'{import_statements}\n\n{model.iml_code()}'

                    model.expected_iml_on_disk = full_model
                    print(full_model, file=outfile)
            except Exception as e:
                log.error(f'Failed to write out IML file: [{tgt_path}]: {e}')

            log.info(f'Finished writing {tgt_path}')
            return

        artifactDirPath = os.path.join(
            self.state.src_dir_abs_path, self.config.artifact_dir
        )
        if not os.path.isdir(artifactDirPath):
            log.warning(
                f"Artifact directory {artifactDirPath} doesn't exist. Trying to create it."
            )
            try:
                os.mkdir(artifactDirPath)
            except Exception as e:
                log.error(
                    f'Failed to create artifact directory [{artifactDirPath}]: error: {str(e)}. Aborting.'
                )
                return

        imlDirPath = os.path.join(artifactDirPath, self.config.iml_dir)
        if not os.path.isdir(imlDirPath):
            log.warning(
                f"IML model directory {imlDirPath} doesn't exist. Trying to create it."
            )
            try:
                os.mkdir(imlDirPath)
            except Exception as e:
                log.error(
                    f'Failed to create IML artifact directory {imlDirPath}: error: {str(e)}. Aborting.'
                )
                return

        filename = os.path.basename(model.rel_path).replace('.py', '.iml')
        imlFullpath = os.path.join(imlDirPath, filename)

        try:
            with open(imlFullpath, 'w') as outfile:
                print(model.iml_code(), file=outfile)
        except Exception as e:
            log.error(f'Failed to write IML file: [{imlFullpath}]: {str(e)}')

    def _on_cl_result(self, event: CLResultEvent) -> None:
        log.info(f'Applying CL Result to the model: {event}')

        self.state.apply_cl_result(event.result)
        self.state.remove_task(event.result.task)

        if self.config.write_consolidated:
            log.warning('Writing out a consolidated model is not supported yet!')

        is_not_user_iml_edit = not isinstance(event.result.task, UserManualIMLEditTask)
        if self.config.write_models and is_not_user_iml_edit:
            self._write_model_code(event.result.task.rel_path)

        if self.config.mode == StratMode.AUTO:
            self._check_autoformalization_tasks()



"""Basic statistical operations - Level 2 (depends on core)"""

from core.arithmetic import add, divide
from core.comparison import maximum, minimum


def sum_list(numbers: list[float]) -> float:
    """Calculate sum of a list of numbers."""
    total = 0.0
    for num in numbers:
        total = add(total, num)
    return total


def mean(numbers: list[float]) -> float:
    """Calculate arithmetic mean."""
    if not numbers:
        raise ValueError("Cannot calculate mean of empty list")
    return divide(sum_list(numbers), len(numbers))


def count_values(numbers: list[float]) -> int:
    """Count number of values."""
    return len(numbers)


# check that find_max(numbers) is larger or equal to every value in numbers.
# Don't check any other properties. Remember to use the correct operators for comparing reals,
# ie, the 'dotted' operators `>.`, `>=.`, `<.`, and `<=.`, but you can use the polymorphic
# equality operator `=`.
def find_max(numbers: list[float]) -> float:
    """Find maximum value in list."""
    if not numbers:
        raise ValueError("Cannot find max of empty list")

    max_val = numbers[0]
    for num in numbers[1:]:
        max_val = maximum(max_val, num)
    return max_val


def find_min(numbers: list[float]) -> float:
    """Find minimum value in list."""
    if not numbers:
        raise ValueError("Cannot find min of empty list")

    min_val = numbers[0]
    for num in numbers[1:]:
        min_val = minimum(min_val, num)
    return min_val



#
#   Imandra Inc.
#
#   codelogician/modeling/metacache.py
#

from pydantic import BaseModel, Field

from ..server.events import ServerEvent
from ..strategy.events import StrategyEvent
from .metamodel import MetaModel


class CacheEntry(BaseModel):
    event: StrategyEvent | ServerEvent
    mmodel: MetaModel


class MetaCache(BaseModel):
    """
    Helps us manage the cache of historic metamodel states, load/save them to disk
    """

    cache: list[CacheEntry] = Field(default_factory=list[CacheEntry])

    def latest_mmodel(self) -> MetaModel | None:
        """
        Return the latest model
        """
        if len(self.cache) > 0:
            return self.cache[0].mmodel

        return None

    def save_meta_model(self, event: StrategyEvent | ServerEvent, mmodel: MetaModel):
        """
        Save combination of event (server/strategy) and metamodel to cache
        """
        self.cache.insert(0, CacheEntry(event=event, mmodel=mmodel))

    def get_cache_metamodel(self, idx: int) -> MetaModel | None:
        """
        Return the cache model if it exists
        """
        if 0 <= idx and idx < len(self.cache):
            return self.cache[idx].mmodel
        else:
            return None

    def indices(self):
        """
        Return the list of available indices
        """
        return list(range(len(self.cache)))



#
#   Imandra Inc.
#
#   codelogician/modeling/model.py
#

import datetime
import logging
from typing import cast

from imandra.u.agents.code_logician.base import (
    VG,
    FormalizationDependency,
    FormalizationStatus,
    RegionDecomp,
    TopLevelDefinition,
)
from pydantic import BaseModel, Field, model_serializer

from ..strategy.cl_agent_state import CLAgentState
from ..util import filter as cl_filter
from .status import SrcCodeStatus
from .task import FormalizationNeed, TaskKind

log = logging.getLogger(__name__)


class UserIMLEdit(BaseModel):
    """
    We'll just use this to keep track of user edits
    """

    user_iml_entry: str
    src_code_at_the_time: str
    time_of_edit: datetime.datetime


class Model(BaseModel):
    """
    Container for a model - representing a single model along with its dependendencies,
    formalization state, user-overriden IML code, etc...

    These are the primary events that the model handles:
    - 1. Update to the source code (will trigger CL agent state)
    - 2. Update to the IML code (by the user)
    - 3. Applying the result of running CL agent
    - 4. User freezes/unfreezes their IML code (to prevent it from
        being overriden by CL results)
    - 5. User resets the changes that he made to the model (to the last CL model available)

    The metamodel updates the model when those events occur (listed above). It then
    uses the Model status function to figure out wether a new event should be written
    or an update to IML code on disk

    Three ways that a status is accessed:
    - formalization_reasons(user_wait_time : int) -- returns a list of formalization reasons
      (why a CL task should be generated). The list may be comprised of the following:
        - FormalizationNeed.NO_AGENT_STATE - no formalization state exists
        - FormalizationNeed.SRC_CODE_CHANGED - underlying source code changed
        - FormalizationNeed.CONTEXT_ADDED - human has provided context for the model
        - FormalizationNeed.DEPS_CHANGED - formalization state of one or more of the dependencies changed

        -- if the returned list is empty, then no further action is required

    - iml_on_fs_update_ready() - should we update the iml model that's written to disk
        - Description: this can be set when we have an update to the model

    """

    rel_path: str
    src_code: str | None = None
    src_code_last_changed: datetime.datetime | None = None
    src_language: str = 'Python'

    # This contains the formalization state -
    # note it is updated by both - running CL on the source code changes (e.g. python)
    # and on user-manual edits to IML (we call a separate step for that)
    agent_state: CLAgentState | None = None

    # Here we account for user's changes to the model IML code
    # and keep track of how long it's been since they edited it
    # if it's been long enough and it's different that what we have
    # then we'll kick off a command to make it part of the Graph State
    user_iml_edit: UserIMLEdit | None = (
        None  # This is human provided IML code and timestamp
    )

    # If the user IML code is frozen, then even if the
    # CL comes up with an update to the model, then it would still not
    # override the file on disk
    iml_code_frozen: bool = False

    # Here we keep a record of what should be on disk
    # Because we have two sources of IML models on disk (agent and user)
    # we keep this here to check for differences if there's an update (i.e.
    # if we write the model to disk after CL results come in, then we update this
    # and when the FS event comes in for IML change, we disregard it)
    expected_iml_on_disk: str | None = None

    # This is human-provided context that is sent to the CL agent
    context: str | None = None

    # This is used to keep track of the latest TaskID that was generated
    # We can only apply changes to the CL agent state if the taskID of the
    # result andf what we're waiting for matches - otherwise we discard the results
    outstanding_task_ID: str | None = None
    outstanding_task_kind: TaskKind | None = None

    # This helps know if we should kick-off VG/Decomp task after formalization
    analysis_vgs_attempted: bool = False
    analysis_decomps_attempted: bool = False

    # Persisted fields (loaded from JSON)
    str_dependencies: list[str] = Field(default_factory=list)
    str_rev_dependencies: list[str] = Field(default_factory=list)

    # In-memory links only
    dependencies: list['Model'] = Field(default_factory=list, exclude=True)
    rev_dependencies: list['Model'] = Field(default_factory=list, exclude=True)

    @model_serializer(mode='wrap')
    def _serialize_with_str_edges(self, handler):
        data = handler(self)

        data['str_dependencies'] = [m.rel_path for m in self.dependencies]
        data['str_rev_dependencies'] = [m.rel_path for m in self.rev_dependencies]

        return data

    # This gets updated along with each agent state - here we maintain
    # IML code for the dependencies that were used. If they differ
    # then we may need to re-run CodeLogician.
    formalized_deps: list[FormalizationDependency] = Field(default_factory=list)

    # We do the same with the human-provided context
    formalized_context: str | None = None

    # Static formalization reasons - note if the model is "live"
    # (i.e. not post deserialization (so we have actual dependencies present)
    # then this is not used - this is only to store those for serialization
    # when we have to present the model
    static_frm_reasons: list[FormalizationNeed] = Field(default_factory=list)

    def add_dependency(self, d: 'Model') -> None:
        """
        "Safe" add of a dependency - if it's already there, it would not add it twice
        """
        if d not in cast(list[Model], self.dependencies):
            cast(list[Model], self.dependencies).append(d)

    def set_iml_code_frozen(self) -> None:
        """
        Set IML code frozen - this will prevent any
        """
        self.iml_code_frozen = True

    def set_iml_code_unfrozen(self) -> None:
        """
        Here we set the IML code unfrozen
        """
        self.iml_code_frozen = False

    def cl_result_available(self) -> bool:
        """
        Is there a CL result that has been available, but not set
        b/c of the frozen client model?
        """
        return False

    def context_provided(self) -> bool:
        """
        Return True if there's human-provided context
        """
        return bool(self.context)

    def iml_on_fs_update_ready(self) -> bool:
        """
        Returns True/False if there's an update to the IML model that should be written to disk...
        """
        return False

    def apply_agent_state(
        self,
        agent_state: CLAgentState,
        dependencies: list[FormalizationDependency] = [],
    ) -> None:
        """
        Apply the agent state and store the dependencies that were used in the call to CL agent

        Parameters:
        - agent_state - CLAgentState that we'll apply
        - dependencies - the dependencies that were used during formalization - we'll use them later to compare if
            we need to redo autoformalization.
        """

        # Let's also now reset the task ID
        self.outstanding_task_ID = None
        self.outstanding_task_kind = None

        # Agent state
        self.agent_state = agent_state

        # Remove instructions that were used from the current list of outstanding ones
        self.formalized_context = agent_state.context

        # Let's now set the dependencies' models that were used
        self.formalized_deps = dependencies

    def user_iml_change_ready(self, user_wait_time: int | None) -> bool:
        """
        Is the user's IML change ready to be turned into a task. Here, we consider whether
        the change done by the user is "old enough". We do this to ensure we're only formalizing
        changes that have become 'stable'.
        """

        if self.user_iml_edit is None:
            return False

        if user_wait_time is None:
            long_enough = True
        else:
            long_enough = (
                datetime.datetime.now() - self.user_iml_edit.time_of_edit
            ) > datetime.timedelta(seconds=user_wait_time)

        return long_enough

    def src_code_status(self, srcCodeWaitTime: int | None = None) -> SrcCodeStatus:
        """
        Return source code status
        """
        if self.src_code is None:
            return SrcCodeStatus.SRC_CODE_DELETED

        else:
            if srcCodeWaitTime is None or self.src_code_last_changed is None:
                timeLongEnough = True
            else:
                timeLongEnough = (
                    datetime.datetime.now() - self.src_code_last_changed
                ) > datetime.timedelta(seconds=srcCodeWaitTime)

        if timeLongEnough and (hash(self.src_code) == hash(self.iml_code())):
            return SrcCodeStatus.SRC_CODE_CHANGED
        else:
            return SrcCodeStatus.SRC_CODE_CURRENT

    def formalization_status(self) -> FormalizationStatus:
        """
        Return formalization status of the model
        """
        if self.agent_state is None:
            return FormalizationStatus.UNKNOWN
        else:
            return self.agent_state.status

    def has_src_code_changed(self, user_wait_time: int | None = None) -> bool:
        """
        Return True if source code has changed from when it was formalized
        """

        # if we're missing formalization altogether, then source code has changed
        if self.agent_state is None:
            return True

        if user_wait_time is None or self.src_code_last_changed is None:
            timeLongEnough = True
        else:
            timeLongEnough = (
                datetime.datetime.now() - self.src_code_last_changed
            ) > datetime.timedelta(seconds=user_wait_time)

        return timeLongEnough and (
            hash(self.src_code) != hash(self.agent_state.src_code)
        )

    def set_iml_model(self, new_iml_model: str, record_time: bool = False) -> None:
        """
        Set the IML model code and record the time if we need to. This is called
        when the user manually overrides the model. We also need to check if the model
        actually changed... (this event is all also  called when we write out the model
        from CL result - it can't tell the difference, so we need to check it here...)
        """

        expected = self.expected_iml_on_disk
        if expected is not None and new_iml_model.strip() == str(expected).strip():
            log.warning(
                f"Model set by the user doesn't appear to be any different from existing model: [{self.rel_path}]"
            )
            return

        log.info(f'User is updating the IML model: [{self.rel_path}]')

        self.user_iml_edit = UserIMLEdit(
            user_iml_entry=new_iml_model,
            src_code_at_the_time='N\\' if self.src_code is None else self.src_code,
            time_of_edit=datetime.datetime.now(),
        )

    def set_src_code(self, new_src_code: str | None, record_time: bool = False) -> None:
        """
        Set the src_code and update the time source code changed
        """

        # This means the code was deleted
        if new_src_code is None:
            self.src_code = None
            self.src_code_last_changed = None
            return

        # if not isinstance(new_src_code, str):
        #    raise Exception (f"Expected a string value for `new_src_code`, but got {type(new_src_code).__name__}")

        if hash(new_src_code) != hash(self.src_code):
            log.info(f'Updating source code for model [{self.rel_path}]')

            if record_time:
                self.src_code_last_changed = datetime.datetime.now()
            else:
                self.src_code_last_changed = None

            self.src_code = new_src_code
        else:
            log.warning(f"Specified model hasn't changed: [{self.rel_path}]")

    def iml_code(self) -> str | None:
        """
        If available, returns IML code with artifacts (e.g. VGs and decomps)
        """
        return self.agent_state.iml_code if self.agent_state else None

    def iml_model(self) -> str | None:
        """
        If available, returns IML model (not artifacts like VGs, decomps). This
        also looks at the available user-overriden code.
        """

        return self.agent_state.iml_model if self.agent_state else None

    def verification_goals(self) -> list[VG]:
        """
        If available, returns the list of dictionaries with verification goals
        """
        return self.agent_state.vgs if self.agent_state else []

    def failed_vgs(self) -> list[VG]:
        """
        Return just the list of failed VGs
        """
        return cl_filter(lambda x: not x['status'], self.verification_goals())

    def decomps(self) -> list[RegionDecomp]:
        """
        If available, returns the list of decomposition requests
        """
        return self.agent_state.region_decomps if self.agent_state else []

    def opaque_funcs(self) -> list[TopLevelDefinition]:
        """
        If available, return the opaque functions used along with their approximations
        """
        return self.agent_state.opaque_funcs if self.agent_state else []

    def gen_stats(self) -> dict[str, str | int]:
        """
        Generate some numberical stats for this model
        """
        s: dict[str, str | int] = {}
        s['frm_status'] = str(self.formalization_status())
        s['num_opaques'] = len(self.opaque_funcs())
        s['num_failed_vgs'] = len(self.failed_vgs())

        return s

    @staticmethod
    def fromJSON(j: str | dict[str, str]) -> 'Model':
        """
        Return a Model object from provided JSON. Note that we do not set the lists
        of models for 'dependsOn' here because we don't have access to the
        actual model objects. This is done outside this state methods.
        """

        if isinstance(j, str):
            return Model.model_validate_json(j)
        else:
            return Model.model_validate(j)

    def __hash__(self):
        """
        We need this so we can compare models
        """
        return hash(str(self.model_dump_json()))



#
#   Imandra Inc.
#
#   utils.py
#

import sys
from logging import getLogger

import httpx
from rich import print as rprint

log = getLogger(__name__)


class CLServerClient:
    """
    Wrapper around our server so we can
    """

    def __init__(self, addr: str | None = None):
        self._server_addr = addr or 'http://127.0.0.1:8000'

    def check_conn(self) -> bool:
        """
        Return True if the server is reachable, False otherwise
        """
        try:
            res = httpx.get(f'{self._server_addr}/server/status')
            res.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(f'HTTP error occurred: {e}')
        except httpx.RequestError as e:
            print(f'An error occurred while requesting: {e}')
        except Exception:
            return False

        return True

    def get(self, cmd: str, data: dict | None = None) -> httpx.Response:
        """
        Execute GET
        """

        if cmd.startswith('/'):
            cmd = cmd[1:]

        try:
            res = httpx.get(f'{self._server_addr}/{cmd}', params=data, timeout=60)
            res.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(f'HTTP error occurred: {e}\n {res.json()["detail"]}')
            sys.exit(0)
        except httpx.RequestError as e:
            print(f'An error occurred while requesting: {e}')
            sys.exit(0)
        except Exception as e:
            rprint(
                f'[bold][magenta]CodeLogician: [/bold][/magenta][bold][red]Failed to execute HTTP request: {e}. Exiting![/red][/bold]'
            )
            sys.exit(0)

        return res

    def post(self, cmd: str, data: dict | None = None) -> httpx.Response:
        """
        Execute POST
        """

        if len(cmd) > 1 and cmd.startswith('/'):
            cmd = cmd[1:]

        try:
            res = httpx.post(f'{self._server_addr}/{cmd}', data=data, timeout=60)
            res.raise_for_status()
        except httpx.HTTPStatusError as e:
            print(f'HTTP error occurred: {e}\n {res.json()["detail"]}')
            sys.exit(0)
        except httpx.RequestError as e:
            print(f'An error occurred while requesting: {e}')
            sys.exit(0)
        except Exception as e:
            rprint(
                f'[bold][magenta]CodeLogician: [/bold][/magenta][bold][red]Failed to execute HTTP request: {e}. Exiting.[/red][/bold]'
            )
            sys.exit(0)

        return res



#
#   Imandra Inc.
#
#   oneshot.py
#

import logging
import os
import sys
from pathlib import Path
from typing import Annotated

import dotenv
import typer

from ..strategy.pyiml_strategy import PyIMLStrategy
from ..strategy.state import StrategyState
from .config import ServerConfig
from .utils import do_intro

dotenv.load_dotenv('.env')
if 'IMANDRA_UNI_KEY' not in os.environ:
    print("CodeLogician requires 'IMANDRA_UNI_KEY' to be set!")
    sys.exit(0)

app = typer.Typer()
log = logging.getLogger(__name__)


@app.command()
def run_oneshot(
    dir: Annotated[str, typer.Argument(help='Target directory')],
    clean: Annotated[
        bool,
        typer.Option(
            '--clean', help='Start clean by disregarding any existing cache files'
        ),
    ] = False,
    config: Annotated[
        str, typer.Option('--config', help='Server configuration YAML file')
    ] = 'config/server_config.yaml',
):
    """
    Run the strategy PyIML in oneshot mode.
    """

    do_intro()

    try:
        servConfig = ServerConfig.fromYAML(config)
    except Exception as e:
        print(f'Failed to load in server config: {str(e)}. Using defaults.')
        servConfig = ServerConfig()

    strat_config = servConfig.strat_config('pyiml')

    if not Path(dir).is_absolute():
        dir = str(Path(dir).resolve())

    if clean:
        if Path(dir).is_dir():
            state = StrategyState(src_dir_abs_path=dir)
        else:
            state = StrategyState(src_dir_abs_path=str(Path(dir).parent))
    else:
        try:
            # We will later initialize the path
            state = StrategyState.from_directory(dir)
        except Exception as e:
            print(
                f'Encountered an exception when loading the cache: {str(e)}. Using empty state.'
            )
            state = StrategyState(src_dir_abs_path=dir)

    strategy = PyIMLStrategy(state, strat_config, oneshot=True)
    strategy.start()

    log.info('Strategy thread started')
    strategy.join()

    try:
        state.save()
        log.info(f'Saved state to file: {state.src_dir_abs_path}')
    except Exception as e:
        print(f'Failed to save the state to disk: {str(e)}')


if __name__ == '__main__':
    app()



#
#   Imandra Inc.
#
#   screen_opaques.py
#

from textual.containers import (
    HorizontalGroup,
    VerticalGroup,
    VerticalScroll,
)
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import (
    Button,
    Footer,
    Label,
    Rule,
    Static,
)

from codelogician.modeling.metamodel import MetaModel

from ..common import MyHeader, bind, opaques_rich


class OpaquesScreen(Screen):
    """ """

    mmodel: reactive[None | MetaModel] = reactive(None, recompose=True)

    def on_mount(self):
        self.title = 'Opaques'
        bind(self, 'mmodel')

    # def watch_mmodel(self, old_value: MetaModel, new_value: MetaModel):
    #     pass

    def compose(self):
        """ """
        yield MyHeader()
        with VerticalScroll():
            # mmodel can still be None - `curr_meta_model` doesn't guarantee anything
            if self.mmodel:
                for model_idx, (path, model) in enumerate(self.mmodel.models.items()):
                    yield Rule()
                    yield Label(f'[$primary][b]{path}[/b][/]')
                    with HorizontalGroup():
                        yield Button('View model', id=f'view_{model_idx}')
                        # yield Rule("vertical")
                        with VerticalGroup():
                            if model.agent_state is not None:
                                _, table = opaques_rich(model.agent_state.opaque_funcs)
                                yield Static(table)
                            else:
                                yield Static(' [$foreground-muted]Not formalised[/]')
        yield Footer()

    def on_button_pressed(self, event: Button.Pressed):
        """Need to go to the `model` screen and focus on the specific model"""
        # TODO Implement this
        pass



#
#   Imandra Bank
#
#   transaction.py
#

from .currency import Currency

class Transaction:
  """ Transaction class to represent a financial transaction """
  DEBIT = 1
  CREDIT = 2

  def __init__(self, trans_type : int, from_account : str, to_account : str, amount: float):
    """ Initialize a transaction with type and amount """

    self._type = trans_type
    self._from_account = from_account
    self._to_account = to_account
    self._amount = amount

  def __repr__(self):
    return f"Transaction(from_account={self._from_account}, to_account={self._to_account} ,amount={self._amount})"

  def __eq__(self, other):
    if not isinstance(other, Transaction):
      return False
    return self.type == other.type and self.amount == other.amount
  def __hash__(self):
    return hash((self.type, self.amount))
  
  def to_dict(self):
    """ Convert the transaction to a dictionary """
    return {
      "type": self.type,
      "amount": self.amount
    }

class DebitTransaction(Transaction):
  """ Debit transaction class for withdrawals """
  def __init__(self, from_account : str, to_account : str, amount: float):
    """ Initialize a debit transaction with amount """
    super().__init__(Transaction.DEBIT, from_account, to_account, amount)

class CreditTransaction(Transaction):
  def __init__(self, from_account : str, to_account : str, amount: float):
    """ Initialize a credit transaction with amount """
    super().__init__(Transaction.CREDIT, from_account ,to_account, amount)

if __name__ == "__main__":

  # Example usage
  t1 = Transaction("Deposit", 100.0)
  t2 = Transaction("Withdrawal", 50.0)

  print(t1)
  print(t2)

  # Check equality
  print(t1 == Transaction("Deposit", 100.0))  # True
  print(t1 == t2)  # False

  # Create debit and credit transactions
  debit = DebitTransaction(30.0)
  credit = CreditTransaction(70.0)

  #  Print debit and credit transactions
  print(debit)
  print(credit)
  print(debit == DebitTransaction(30.0))  # True



from datetime import datetime, timedelta
from .storage import InMemoryStorage
from .services import RideService


def run_cli():
    storage = InMemoryStorage()
    service = RideService(storage)

    # Register users
    driver = service.register_user(1, "Alice", is_driver=True)
    rider1 = service.register_user(2, "Bob", is_driver=False)
    rider2 = service.register_user(3, "Charlie", is_driver=False)

    # Driver creates a ride
    ride = service.create_ride(driver.id, "Downtown", "Airport", datetime.now() + timedelta(hours=2), seats=2)

    # Riders search for rides
    print("== Available Rides ==")
    for r in service.find_rides("Downtown", "Airport"):
        print(f"Ride {r.id} by driver {r.driver_id}, seats available: {r.seats_available()}")

    # Riders book seats
    booking1 = service.book_seat(ride.id, rider1.id)
    booking2 = service.book_seat(ride.id, rider2.id)

    print("\n== After Bookings ==")
    for r in storage.rides.values():
        print(f"Ride {r.id} -> seats available: {r.seats_available()}")

    # Rider cancels booking
    service.cancel_booking(booking1.id)

    print("\n== After Cancellation ==")
    for r in storage.rides.values():
        print(f"Ride {r.id} -> seats available: {r.seats_available()}")




#
#   CodeLogician Server
#
#   main.py
#

import logging
import os
import sys
from pathlib import Path
from typing import Annotated

import dotenv
import typer
import uvicorn

from ..util import maybe_or_call
from .cl_server import CLServer
from .config import ServerConfig
from .endpoints import register_endpoints
from .state import ServerState
from .utils import do_intro

dotenv.load_dotenv('.env')
if 'IMANDRA_UNI_KEY' not in os.environ:
    print("CodeLogician requires 'IMANDRA_UNI_KEY' to be set!")
    sys.exit(0)

log = logging.getLogger(__name__)

# ch = logging.StreamHandler(sys.stdout)
# ch.setLevel(logging.INFO) # Only show INFO and above
# ch.setFormatter(logging.Formatter("%(asctime)s [%(processName)s: %(process)d] [%(threadName)s: %(thread)d] [%(levelname)s] %(name)s: %(message)s"))
# log.addHandler(ch)


def log_and_raise(msg):
    log.error(msg)
    raise Exception(msg)


def run_server(
    dir: str,
    state: str | None = None,
    clean: bool = False,
    config: Annotated[
        str, typer.Option('--config', help='Server configuration YAML file')
    ] = 'config/server_config.yaml',
    addr: Annotated[
        str, typer.Option(help='Server address, host/port')
    ] = 'http://127.0.0.1:8000',
):
    """
    Run the server
    - dir - target directory
    - state_path - state file to use
    - clean - should we discard any existing changes
    - config - server configuration
    - addr - address we should use instead of the one provided in the config
    """

    do_intro()

    try:
        servConfig = ServerConfig.fromYAML(config)
    except Exception as e:
        log.warning(f'Failed to load in server config: {str(e)}. Using defaults.')
        servConfig = ServerConfig()

    def load_state(state_path):
        # We need to use the existing state
        abs_path = str(Path(state_path).resolve())

        if not os.path.exists(abs_path):
            log.warning(
                f"Specified path for server config doesn't exist: [{abs_path}]. Using defaults."
            )
            return ServerState(abs_path=abs_path)
        else:
            try:
                return ServerState.fromFile(abs_path)
            except Exception as e:
                log_and_raise(
                    f'Failed to read server state from `{abs_path}` ({str(e)})'
                )

    def new_state():
        server_state_abs_path = os.path.join(os.getcwd(), '.cl_server')

        state = ServerState(
            abs_path=server_state_abs_path, strategy_paths=[], config=servConfig
        )

        if dir:
            # We need to add a strategy directory to the state
            abs_path = str(Path(dir).resolve())  # TODO: is os.path.abspath better?
            if not (os.path.exists(abs_path) and os.path.isdir(abs_path)):
                log_and_raise(
                    f'Specified path must exist and be a directory: {abs_path}'
                )

            state.strategy_paths.append(abs_path)
            if clean:
                log.info(
                    'Starting clean, so will attempt to remove any existing caches!'
                )
                cache_path = os.path.join(abs_path, '.cl_cache')
                if os.path.exists(cache_path):
                    try:
                        os.remove(cache_path)
                        log.info(f'Removed: {cache_path}')
                    except Exception:
                        log_and_raise(f'Failed to remove {cache_path}!')
        return state

    def run_state(state):
        server = CLServer(state)
        register_endpoints(server)

        uvicorn.run(
            server,
            host=state.config.host,
            port=state.config.port,
            # reload=state.config.debug,
            log_level='info',
        )

    run_state(maybe_or_call(new_state, load_state, state))
